############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging

from .stats import EnumStats
from .long_read_counter import (
    ExonCounter,
    IntronCounter,
    CompositeCounter,
    create_gene_counter,
    create_transcript_counter,
)
from .assignment_io import (
    IOSupport,
    BEDPrinter,
    ReadAssignmentCompositePrinter,
    SqantiTSVPrinter,
    BasicTSVAssignmentPrinter,
)
from .transcript_printer import VoidTranscriptPrinter
from .gene_info import get_all_chromosome_genes, get_all_chromosome_transcripts
from .common import large_output_enabled

logger = logging.getLogger('IsoQuant')


class ReadAssignmentAggregator:
    def __init__(self, args, sample, string_pools, gffutils_db=None, chr_id=None, gzipped=False, grouping_strategy_names=None):
        self.args = args
        self.string_pools = string_pools
        self.grouping_strategy_names = grouping_strategy_names if grouping_strategy_names else ["default"]
        self.common_header = "# Command line: " + args._cmd_line + "\n# IsoQuant version: " + args._version + "\n"
        self.io_support = IOSupport(self.args)

        self.gene_set = set()
        self.transcript_set = set()
        if gffutils_db and chr_id:
            self.gene_set = set(get_all_chromosome_genes(gffutils_db, chr_id))
            self.transcript_set = set(get_all_chromosome_transcripts(gffutils_db, chr_id))

        self.read_stat_counter = EnumStats()

        printer_list = []
        self.corrected_bed_printer = None
        if large_output_enabled(self.args, "corrected_bed"):
            corrected_bed_path = sample.get_corrected_bed_file(chr_id) if chr_id else sample.out_corrected_bed
            self.corrected_bed_printer = BEDPrinter(corrected_bed_path,
                                                    self.args,
                                                    print_corrected=True,
                                                    gzipped=gzipped)
            printer_list.append(self.corrected_bed_printer)
        self.basic_printer = None
        if self.args.genedb and large_output_enabled(self.args, "read_assignments"):
            assigned_tsv_path = sample.get_assigned_tsv_file(chr_id) if chr_id else sample.out_assigned_tsv
            self.basic_printer = BasicTSVAssignmentPrinter(assigned_tsv_path, self.args, self.io_support,
                                                           additional_header=self.common_header, gzipped=gzipped)
            sample.out_assigned_tsv_result = self.basic_printer.output_file_name
            printer_list.append(self.basic_printer)
        self.t2t_sqanti_printer = VoidTranscriptPrinter()
        if self.args.sqanti_output:
            t2t_path = sample.get_t2t_tsv_file(chr_id) if chr_id else sample.out_t2t_tsv
            self.t2t_sqanti_printer = SqantiTSVPrinter(t2t_path, self.args, self.io_support)
        self.global_printer = ReadAssignmentCompositePrinter(printer_list)

        self.global_counter = CompositeCounter()
        if self.args.genedb:
            gene_counts_path = sample.get_gene_counts_file(chr_id) if chr_id else sample.out_gene_counts_tsv
            transcript_counts_path = sample.get_transcript_counts_file(chr_id) if chr_id else sample.out_transcript_counts_tsv
            self.gene_counter = create_gene_counter(gene_counts_path,
                                                    self.args.gene_quantification,
                                                    complete_feature_list=self.gene_set)
            self.transcript_counter = create_transcript_counter(transcript_counts_path,
                                                                self.args.transcript_quantification,
                                                                complete_feature_list=self.transcript_set)
            self.global_counter.add_counters([self.gene_counter, self.transcript_counter])

        self.transcript_model_global_counter = CompositeCounter()
        self.gene_model_global_counter = CompositeCounter()
        if not self.args.no_model_construction:
            transcript_model_counts_path = sample.get_transcript_model_counts_file(chr_id) if chr_id else sample.out_transcript_model_counts_tsv
            gene_model_counts_path = sample.get_gene_model_counts_file(chr_id) if chr_id else sample.out_gene_model_counts_tsv
            self.transcript_model_counter = create_transcript_counter(transcript_model_counts_path,
                                                                      self.args.transcript_quantification)
            self.gene_model_counter = create_gene_counter(gene_model_counts_path,
                                                          self.args.gene_quantification)

            self.transcript_model_global_counter.add_counter(self.transcript_model_counter)
            self.gene_model_global_counter.add_counter(self.gene_model_counter)

        if self.args.count_exons and self.args.genedb:
            exon_counts_path = sample.get_exon_counts_file(chr_id) if chr_id else sample.out_exon_counts_tsv
            intron_counts_path = sample.get_intron_counts_file(chr_id) if chr_id else sample.out_intron_counts_tsv
            # string_pools=None means ungrouped counting
            self.exon_counter = ExonCounter(exon_counts_path)
            self.intron_counter = IntronCounter(intron_counts_path)
            self.global_counter.add_counters([self.exon_counter, self.intron_counter])

        if self.args.read_group and self.args.genedb:
            for group_idx, strategy_name in enumerate(self.grouping_strategy_names):
                # Use chr-specific paths if chr_id is provided
                if chr_id:
                    gene_out_file = sample.get_grouped_counts_file(chr_id, "gene", strategy_name)
                    transcript_out_file = sample.get_grouped_counts_file(chr_id, "transcript", strategy_name)
                else:
                    gene_out_file = f"{sample.out_gene_grouped_counts_tsv}_{strategy_name}"
                    transcript_out_file = f"{sample.out_transcript_grouped_counts_tsv}_{strategy_name}"

                gene_counter = create_gene_counter(gene_out_file,
                                                   self.args.gene_quantification,
                                                   complete_feature_list=self.gene_set,
                                                   string_pools=self.string_pools,
                                                   group_index=group_idx)
                transcript_counter = create_transcript_counter(transcript_out_file,
                                                              self.args.transcript_quantification,
                                                              complete_feature_list=self.transcript_set,
                                                              string_pools=self.string_pools,
                                                              group_index=group_idx)

                self.global_counter.add_counters([gene_counter, transcript_counter])

                if self.args.count_exons:
                    if chr_id:
                        exon_out_file = sample.get_grouped_counts_file(chr_id, "exon", strategy_name)
                        intron_out_file = sample.get_grouped_counts_file(chr_id, "intron", strategy_name)
                    else:
                        exon_out_file = f"{sample.out_exon_grouped_counts_tsv}_{strategy_name}"
                        intron_out_file = f"{sample.out_intron_grouped_counts_tsv}_{strategy_name}"
                    exon_counter = ExonCounter(exon_out_file, string_pools=self.string_pools, group_index=group_idx)
                    intron_counter = IntronCounter(intron_out_file, string_pools=self.string_pools, group_index=group_idx)
                    self.global_counter.add_counters([exon_counter, intron_counter])

        if self.args.read_group and not self.args.no_model_construction:
            for group_idx, strategy_name in enumerate(self.grouping_strategy_names):
                if chr_id:
                    transcript_model_out_file = sample.get_grouped_counts_file(chr_id, "discovered_transcript", strategy_name)
                    gene_model_out_file = sample.get_grouped_counts_file(chr_id, "discovered_gene", strategy_name)
                else:
                    transcript_model_out_file = f"{sample.out_transcript_model_grouped_counts_tsv}_{strategy_name}"
                    gene_model_out_file = f"{sample.out_gene_model_grouped_counts_tsv}_{strategy_name}"

                transcript_model_counter = create_transcript_counter(
                    transcript_model_out_file,
                    self.args.transcript_quantification,
                    string_pools=self.string_pools,
                    group_index=group_idx)
                gene_model_counter = create_gene_counter(
                    gene_model_out_file,
                    self.args.gene_quantification,
                    string_pools=self.string_pools,
                    group_index=group_idx)

                self.transcript_model_global_counter.add_counter(transcript_model_counter)
                self.gene_model_global_counter.add_counter(gene_model_counter)



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging

logger = logging.getLogger('IsoQuant')


from .common import reverse_complement


# given a cigar string of an alignment, move shift bases along the alignment from the start
# if shift is negative, moves from the alignment's end
# return value is always positive
def move_ref_coord_alogn_alignment(alignment, shift):
    if shift == 0:
        return 0

    cigar_tuples = alignment.cigartuples
    assert cigar_tuples

    direction = 1 if shift > 0 else -1 # 1 for forward -1 for backward
    shift = abs(shift)

    if direction == 1:
        current_pos = 0
        if len(cigar_tuples) > 1 and cigar_tuples[0][0] == 5 and cigar_tuples[1][0] == 4:
            # hard clipped
            current_pos = 2
        elif cigar_tuples[0][0] in [4, 5]:
            # soft clipped
            current_pos = 1
    else:
        current_pos = -1
        if len(cigar_tuples) > 1 and cigar_tuples[-1][0] == 5 and cigar_tuples[-2][0] == 4:
            # hard clipped
            current_pos -= 2
        elif cigar_tuples[-1][0] in [4, 5]:
            # soft clipped
            current_pos -= 1

    read_length_consumed = 0
    reference_length_consumed = 0
    shift += 1 # a hack to jump to the next event
    while current_pos < len(cigar_tuples) and current_pos >= -len(cigar_tuples) and read_length_consumed < shift:
        cigar_event = cigar_tuples[current_pos][0]
        event_len = cigar_tuples[current_pos][1]
        if cigar_event == 1:
            # insertion
            read_length_consumed += event_len
        elif cigar_event in [2, 3]:
            # deletion or intron
            reference_length_consumed += event_len
        elif cigar_event in [0, 7, 8]:
            # match
            remaining_bases = shift - read_length_consumed # how many nucleotides in read to complete shift
            #logger.debug("%d" % remaining_bases)
            if event_len < remaining_bases:
                reference_length_consumed += event_len
                read_length_consumed += event_len
            else:
                read_length_consumed += remaining_bases
                reference_length_consumed += remaining_bases
        elif cigar_event in [4, 5]:
            # met clipping on the other side
            break
        else:
            # unexpected event
            logger.warning("Unexpected event: " + cigar_event)
        #logger.debug("%d, %d, %d, %d" % (cigar_event, event_len, read_length_consumed, reference_length_consumed))

        current_pos += direction

    return reference_length_consumed-1


class PolyAInfo:
    def __init__(self, external_polya_pos, external_polyt_pos, internal_polya_pos, internal_polyt_pos):
        self.external_polya_pos = external_polya_pos
        self.external_polyt_pos = external_polyt_pos
        self.internal_polya_pos = internal_polya_pos
        self.internal_polyt_pos = internal_polyt_pos


class PolyAFinder:
    def __init__(self, window_size=16, min_polya_fraction=0.75):
        self.window_size = window_size
        self.min_polya_fraction = min_polya_fraction
        self.polyA_count = int(self.window_size * self.min_polya_fraction)

    def detect_polya(self, alignement):
        return PolyAInfo(self.find_polya_external(alignement), self.find_polyt_external(alignement),
                         self.find_polya_internal(alignement), self.find_polyt_internal(alignement))

    def find_polya_external(self, alignment):
        return self.find_polya_tail(alignment, 2, 2 * self.window_size)

    def find_polya_internal(self, alignment):
        return self.find_polya_tail(alignment, 4 * self.window_size, 2, check_entire_tail=True)

    def find_polyt_external(self, alignment):
        return self.find_polyt_head(alignment, 2, 2 * self.window_size)

    def find_polyt_internal(self, alignment):
        return self.find_polyt_head(alignment, 4 * self.window_size, 2, check_entire_head=True)

    # == polyA stuff ==
    def find_polya_tail(self, alignment, from_pos, to_pos, check_entire_tail=False):
        # logger.debug("Detecting polyA tail for %s " % alignment.query_name)
        cigar_tuples = alignment.cigartuples
        soft_clipped_tail_len = 0

        if len(cigar_tuples) > 1 and cigar_tuples[-1][0] == 5 and cigar_tuples[-2][0] == 4:
            # hard clipped
            soft_clipped_tail_len = cigar_tuples[-2][1]
        elif cigar_tuples[-1][0] == 4:
            # soft clipped
            soft_clipped_tail_len = cigar_tuples[-1][1]

        seq = alignment.seq
        if not seq:
            return -1
        assert soft_clipped_tail_len < len(seq)

        read_mapped_region_end = len(seq) - soft_clipped_tail_len
        to_check_start = max(0, read_mapped_region_end - from_pos)
        to_check_end = min(len(seq), read_mapped_region_end + to_pos + 1)
        sequence_to_check = alignment.seq[to_check_start:to_check_end].upper()
        pos = self.find_polya(sequence_to_check)

        #logger.debug("read start: %d, ckeck start: %d, check end: %d, pos: %d" % (read_mapped_region_end, to_check_start, to_check_end, pos))
        #logger.debug(sequence_to_check)
        if check_entire_tail and pos != -1:
            entire_tail = sequence_to_check[pos:]
            if entire_tail.count('A') < len(entire_tail) * self.min_polya_fraction:
                logger.debug("Internal polyA seems unreliable")
                pos = -1

        if pos == -1:
            logger.debug("No polyA found")
            return -1

        # add position of the region we check
        pos = to_check_start + pos
        if pos >= read_mapped_region_end: # poly A starts after last mapped base
            shift = pos - read_mapped_region_end
            reference_polya_start = alignment.reference_end + shift
        else:
            shift = pos - read_mapped_region_end
            ref_shift = move_ref_coord_alogn_alignment(alignment, shift)
            reference_polya_start = alignment.reference_end - ref_shift
            #logger.debug("shift: %d, ref shift: %d, reference: %d" % (shift, ref_shift, reference_polya_start))

        logger.debug("PolyA found at position %d" % reference_polya_start)
        return reference_polya_start

    def find_polyt_head(self, alignment, from_pos, to_pos, check_entire_head=False):
        # logger.debug("Detecting polyT head for %s " % alignment.query_name)
        cigar_tuples = alignment.cigartuples
        soft_clipped_head_len = 0

        if len(cigar_tuples) > 1 and cigar_tuples[0][0] == 5 and cigar_tuples[1][0] == 4:
            # hard clipped
            soft_clipped_head_len = cigar_tuples[1][1]
        elif cigar_tuples[0][0] == 4:
            # soft clipped
            soft_clipped_head_len = cigar_tuples[0][1]

        seq = alignment.seq
        if not seq:
            return -1
        assert soft_clipped_head_len < len(seq)

        read_mapped_region_start = soft_clipped_head_len
        to_check_start = max(0, read_mapped_region_start - to_pos)
        to_check_end = min(len(seq), read_mapped_region_start + from_pos + 1)
        sequence_to_check = reverse_complement(str(alignment.seq[to_check_start:to_check_end])).upper()

        pos = self.find_polya(sequence_to_check)
        #logger.debug("read start: %d, ckeck start: %d, check end: %d, pos: %d" % (read_mapped_region_start, to_check_start, to_check_end, pos))
        #logger.debug(sequence_to_check)

        if check_entire_head and pos != -1:
            entire_tail = sequence_to_check[pos:]
            if entire_tail.count('A') < len(entire_tail) * self.min_polya_fraction:
                #logger.debug("Internal polyT seems unreliable")
                pos = -1

        if pos == -1:
            logger.debug("No polyT found")
            return -1

        # reverse position
        pos = to_check_end - pos - 1
        if pos <= read_mapped_region_start:  # poly T starts to the left of first mapped base
            shift = read_mapped_region_start - pos
            reference_polyt_end = alignment.reference_start - shift
        else:
            shift = pos - read_mapped_region_start
            ref_shift = move_ref_coord_alogn_alignment(alignment, shift)
            reference_polyt_end = alignment.reference_start + ref_shift
            #logger.debug("shift: %d, ref shift: %d, reference: %d" % (shift, ref_shift, reference_polyt_end))

        logger.debug("PolyT found at position %d" % reference_polyt_end)
        return max(1, reference_polyt_end)

    # poly A tail detection
    def find_polya(self, seq):
        if len(seq) < self.window_size:
            return -1
        i = 0
        a_count = seq[0:self.window_size].count('A')
        while i < len(seq) - self.window_size:
            if a_count >= self.polyA_count:
                break
            first_base_a = seq[i] == 'A'
            new_base_a = i + self.window_size < len(seq) and seq[i + self.window_size] == 'A'
            if first_base_a and not new_base_a:
                a_count -= 1
            elif not first_base_a and new_base_a:
                a_count += 1
            i += 1

        if i >= len(seq) - self.window_size:
            return -1

        return i + max(0, seq[i:].find('AA'))




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import json
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.spatial.distance import euclidean


def parse_data(data):
    genes = {}
    for condition, condition_data in data.items():
        for gene, gene_data in condition_data.items():
            if gene not in genes:
                genes[gene] = {
                    "chromosome": gene_data["chromosome"],
                    "start": gene_data["start"],
                    "end": gene_data["end"],
                    "strand": gene_data["strand"],
                    "biotype": gene_data["biotype"],
                    "transcripts": {},
                }
            genes[gene]["transcripts"][condition] = gene_data["transcripts"]
            genes[gene][condition] = gene_data["value"]
    return genes


def calculate_deviance(wt_transcripts, condition_transcripts):
    all_transcripts = set(wt_transcripts.keys()).union(
        set(condition_transcripts.keys())
    )

    wt_proportions = [wt_transcripts.get(t, 0) for t in all_transcripts]
    condition_proportions = [condition_transcripts.get(t, 0) for t in all_transcripts]

    total_wt = sum(wt_proportions)
    total_condition = sum(condition_proportions)

    if total_wt > 0:
        wt_proportions = [p / total_wt for p in wt_proportions]
    if total_condition > 0:
        condition_proportions = [p / total_condition for p in condition_proportions]

    distance = euclidean(wt_proportions, condition_proportions)

    # Reduce distance if total unique transcripts are 1
    if len(all_transcripts) == 1:
        distance *= 0.7

    return distance


def calculate_metrics(genes):
    metrics = []
    for gene, gene_data in genes.items():
        wt_transcripts = gene_data["transcripts"].get("wild_type", {})

        for condition in gene_data:
            if condition in [
                "chromosome",
                "start",
                "end",
                "strand",
                "biotype",
                "transcripts",
                "wild_type",
            ]:
                continue
            condition_transcripts = gene_data["transcripts"].get(condition, {})
            deviance = calculate_deviance(wt_transcripts, condition_transcripts)
            metrics.append({"gene": gene, "condition": condition, "deviance": deviance})

            value = gene_data.get(condition, 0)
            wt_value = gene_data.get("wild_type", 0)
            abs_diff = abs(value - wt_value)
            metrics.append(
                {
                    "gene": gene,
                    "condition": condition,
                    "value": value,
                    "abs_diff": abs_diff,
                }
            )

    return pd.DataFrame(metrics)


def check_known_target(gene, known_targets):
    for target in known_targets:
        if "|" in target:
            if any(part in gene for part in target.split("|")):
                return 1
        elif target == gene:
            return 1
    return 0


def rank_genes(df, known_genes_path=None):
    if known_genes_path:
        target_genes_df = pd.read_csv(known_genes_path, header=None, names=["gene"])
        known_targets = target_genes_df["gene"].tolist()
        df["known_target"] = df["gene"].apply(
            lambda x: check_known_target(x, known_targets)
        )
    else:
        df["known_target"] = 0

    value_ranking = df.groupby("gene")["value"].mean().reset_index()
    abs_diff_ranking = df.groupby("gene")["abs_diff"].mean().reset_index()
    deviance_ranking = df.groupby("gene")["deviance"].mean().reset_index()

    value_ranking["rank_value"] = value_ranking["value"].rank(ascending=False)
    abs_diff_ranking["rank_abs_diff"] = abs_diff_ranking["abs_diff"].rank(
        ascending=False
    )
    deviance_ranking["rank_deviance"] = deviance_ranking["deviance"].rank(
        ascending=False
    )

    merged_df = value_ranking[["gene", "rank_value"]].merge(
        abs_diff_ranking[["gene", "rank_abs_diff"]], on="gene"
    )
    merged_df = merged_df.merge(deviance_ranking[["gene", "rank_deviance"]], on="gene")
    merged_df = merged_df.merge(df[["gene", "known_target"]], on="gene")
    # Devalue the importance of overall expression by reducing its weight
    merged_df["combined_rank"] = (
        merged_df["rank_value"]  # Reduced weight for rank_value
        + merged_df["rank_abs_diff"]
        + merged_df["rank_deviance"]
    )

    top_combined_ranking = merged_df.sort_values(by="combined_rank").head(10)
    top_deviance_ranking = merged_df.sort_values(by="rank_deviance").head(10)
    top_100_combined_ranking = merged_df.sort_values(by="combined_rank").head(100)

    return (
        top_combined_ranking,
        top_deviance_ranking,
        top_100_combined_ranking,
        merged_df,
    )


def visualize_ranking(
    top_combined_ranking, top_deviance_ranking, merged_df, output_dir
):
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Bar plot for combined rank
    plt.figure(figsize=(12, 8))
    sns.barplot(
        x="combined_rank", y="gene", data=top_combined_ranking, palette="viridis"
    )
    plt.title("Top 10 Genes by Combined Ranking")
    plt.xlabel("Combined Rank")
    plt.ylabel("Gene")
    plt.savefig(os.path.join(output_dir, "top_genes_combined_ranking.png"), dpi=300)
    plt.close()

    # Heatmap for metric ranks
    top_genes = top_combined_ranking["gene"].tolist()
    heatmap_data = merged_df[merged_df["gene"].isin(top_genes)]
    heatmap_data = heatmap_data.set_index("gene")[
        ["rank_value", "rank_abs_diff", "rank_deviance"]
    ]

    plt.figure(figsize=(12, 8))
    sns.heatmap(
        heatmap_data,
        annot=True,
        cmap="RdBu_r",
        linewidths=0.5,
        cbar_kws={"label": "Rank"},
    )
    plt.title("Metric Ranks for Top 10 Genes")
    plt.savefig(os.path.join(output_dir, "metric_ranks_heatmap.png"), dpi=300)
    plt.close()

    # Diverging bar plot for deviance
    plt.figure(figsize=(12, 8))
    sns.barplot(
        x="rank_deviance",
        y="gene",
        data=top_deviance_ranking,
        palette="coolwarm",
        orient="h",
    )
    plt.title("Top 10 Genes by Transcript Deviance from Wild Type")
    plt.xlabel("Rank of Deviance from Wild Type")
    plt.ylabel("Gene")
    plt.axvline(x=0, color="grey", linestyle="--")
    plt.savefig(os.path.join(output_dir, "deviance_from_wild_type.png"), dpi=300)
    plt.close()

    # Scatter plot for rank_value vs rank_abs_diff
    plt.figure(figsize=(12, 8))
    sns.scatterplot(
        x="rank_value",
        y="rank_abs_diff",
        hue="gene",
        data=top_deviance_ranking,
        palette="deep",
        s=100,
    )
    plt.title("Rank Value vs Rank Absolute Difference")
    plt.xlabel("Rank Value")
    plt.ylabel("Rank Absolute Difference")
    plt.savefig(os.path.join(output_dir, "rank_value_vs_rank_abs_diff.png"), dpi=300)
    plt.close()

    # Combined multi-metric visualization
    fig, axes = plt.subplots(2, 2, figsize=(20, 16))
    sns.barplot(
        x="combined_rank",
        y="gene",
        data=top_combined_ranking,
        palette="viridis",
        ax=axes[0, 0],
    )
    axes[0, 0].set_title("Combined Rank")
    axes[0, 0].set_xlabel("Combined Rank")
    axes[0, 0].set_ylabel("Gene")

    sns.heatmap(
        heatmap_data,
        annot=True,
        cmap="RdBu_r",
        linewidths=0.5,
        cbar_kws={"label": "Rank"},
        ax=axes[0, 1],
    )
    axes[0, 1].set_title("Metric Ranks")

    sns.barplot(
        x="rank_deviance",
        y="gene",
        data=top_deviance_ranking,
        palette="coolwarm",
        orient="h",
        ax=axes[1, 0],
    )
    axes[1, 0].set_title("Transcript Deviance from Wild Type")
    axes[1, 0].set_xlabel("Rank of Deviance from Wild Type")
    axes[1, 0].set_ylabel("Gene")
    axes[1, 0].axvline(x=0, color="grey", linestyle="--")

    sns.scatterplot(
        x="rank_value",
        y="rank_abs_diff",
        hue="gene",
        data=top_deviance_ranking,
        palette="deep",
        s=100,
        ax=axes[1, 1],
    )
    axes[1, 1].set_title("Rank Value vs Rank Absolute Difference")
    axes[1, 1].set_xlabel("Rank Value")
    axes[1, 1].set_ylabel("Rank Absolute Difference")

    plt.tight_layout()
    plt.savefig(os.path.join(output_dir, "combined_visualization.png"), dpi=300)
    plt.close()


def save_top_genes(top_combined_ranking, output_dir, num_genes):
    top_combined_ranking.head(num_genes)[["gene"]].to_csv(
        os.path.join(output_dir, f"top_{num_genes}_genes.txt"),
        index=False,
        header=False,
        sep="\t",
    )
    return os.path.join(output_dir, f"top_{num_genes}_genes.txt")


def rank_and_visualize_genes(
    input_data, output_dir, num_genes=100, known_genes_path=None
):
    genes = parse_data(input_data)
    metrics_df = calculate_metrics(genes)
    top_combined_ranking, top_deviance_ranking, top_100_combined_ranking, merged_df = (
        rank_genes(metrics_df, known_genes_path)
    )
    merged_df = merged_df.drop_duplicates(subset="gene", keep="first")
    top_combined_ranking = merged_df.sort_values(by="combined_rank").head(num_genes)
    top_deviance_ranking = merged_df.sort_values(by="rank_deviance").head(num_genes)

    visualize_ranking(top_combined_ranking, top_deviance_ranking, merged_df, output_dir)
    path = save_top_genes(top_combined_ranking, output_dir, num_genes)

    print(f"\nTop {num_genes} Genes by Combined Ranking:")
    print(top_combined_ranking[["gene", "combined_rank"]])
    print(f"\nDetailed Metrics for Top {num_genes} Genes by Combined Ranking:")
    print(top_combined_ranking)

    merged_df.to_csv(os.path.join(output_dir, "gene_metrics.csv"), index=False)

    return path



###########################################################################
# Copyright (c) 2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import math
import sys

from ..indexers import ArrayKmerIndexer, Array2BitKmerIndexer, Dict2BitKmerIndexer, KmerIndexer
from ..common import batch_str_to_2bit
from ...error_codes import IsoQuantExitCode
from .extraction_result import DetectedElement, ExtractionResult
from ..common import (
    find_polyt,
    reverese_complement,
    detect_exact_positions,
    find_candidate_with_max_score_ssw,
    find_optimal_kmer_size,
)
from .molecule_structure import ElementType, MoleculeStructure, MoleculeElement

logger = logging.getLogger('IsoQuant')


class UniversalSingleMoleculeExtractor:
    MIN_SCORE_COEFF = 0.75
    MIN_SCORE_COEFF_TERMMINAL = 0.5
    TERMINAL_MATCH_DELTA = 3
    MAX_LEN_DIFF = 0.25
    DEFAULT_ERROR_RATE = 0.03 # rought estimate for modern nanopore, doesn't matter that much
    MAX_HITS = 10
    SCORE_DIFF = 1

    def __init__(self, molecule_structure: MoleculeStructure):
        self.correct_sequences = True
        self.molecule_structure = molecule_structure
        self.index_dict = {}
        self.min_scores = {}
        self.has_polyt = False
        self.has_cdna = False
        self.constant_elements_to_detect = set()
        self.elements_to_extract = set()
        self.elements_to_correct = set()

        for el in self.molecule_structure:
            if el.element_type == ElementType.PolyT:
                if self.has_polyt:
                    logger.critical(
                        "Current version only supports a single polyT, extraction results may be suboptimal")
                    sys.exit(IsoQuantExitCode.INVALID_PARAMETER)
                self.has_polyt = True
            elif el.element_type == ElementType.cDNA:
                if self.has_cdna:
                    logger.critical(
                        "Current version only supports a single cDNA, reads will not be split into molecules")
                    sys.exit(IsoQuantExitCode.INVALID_PARAMETER)
                self.has_cdna = True
            elif el.element_type.is_constant():
                self.constant_elements_to_detect.add(el.element_name)
                self.index_dict[el.element_name] = KmerIndexer([el.element_value],
                                                               max(6, int(el.element_length / 2) - 2))
            elif el.element_type.needs_sequence_extraction():
                self.elements_to_extract.add(el.element_name)
            elif el.element_type.needs_correction():
                if el.element_name in self.molecule_structure.elements_to_concatenate:
                    # Concatenated parts: skip individual index, handled after loop
                    pass
                elif el.element_name in self.molecule_structure.duplicated_elements:
                    # Duplicated parts: skip individual index, handled after loop
                    pass
                elif self.correct_sequences:
                    self.prepare_barcode_index_for_element(el)
                    self.elements_to_correct.add(el.element_name)
                else:
                    self.elements_to_extract.add(el.element_name)

        # Build one index per concatenated base element using the full whitelist
        if self.correct_sequences:
            for base_name, count in self.molecule_structure.concatenated_elements_counts.items():
                first_part_el = None
                total_length = 0
                for el in self.molecule_structure:
                    if el.element_name in self.molecule_structure.elements_to_concatenate:
                        if self.molecule_structure.elements_to_concatenate[el.element_name][0] == base_name:
                            if first_part_el is None:
                                first_part_el = el
                            total_length += el.element_length
                if first_part_el is not None and first_part_el.element_value is not None:
                    barcode_list = first_part_el.element_value
                    self.prepare_barcode_index(base_name, barcode_list, total_length)

        # Build one index per duplicated base element using the full whitelist
        if self.correct_sequences:
            for base_name, count in self.molecule_structure.duplicated_elements_counts.items():
                first_part_el = None
                for el in self.molecule_structure:
                    if el.element_name in self.molecule_structure.duplicated_elements:
                        if self.molecule_structure.duplicated_elements[el.element_name][0] == base_name:
                            first_part_el = el
                            break
                if first_part_el is not None and first_part_el.element_value is not None:
                    barcode_list = first_part_el.element_value
                    self.prepare_barcode_index(base_name, barcode_list, first_part_el.element_length)

        if not self.has_cdna:
            logger.critical("Molecule must include a cDNA")
            sys.exit(IsoQuantExitCode.INVALID_PARAMETER)

    def prepare_barcode_index(self, base_name, barcode_list, barcode_length):
        # Check whether all barcodes have the same length
        barcode_lengths = set(len(b) for b in barcode_list)
        variable_length = len(barcode_lengths) > 1
        if variable_length:
            logger.warning("Barcodes for element %s have variable lengths (%s), "
                           "performance may be suboptimal" %
                           (base_name, ", ".join(str(l) for l in sorted(barcode_lengths))))
            barcode_length = min(barcode_lengths)

        barcode_count = len(barcode_list)
        error_rate = self.DEFAULT_ERROR_RATE
        barcode_sparsity = math.pow(4, barcode_length) / barcode_count
        filling_edit_distance = math.ceil(math.log(barcode_sparsity, 8 * barcode_length))
        error_probability = math.pow(1 - error_rate, barcode_length - filling_edit_distance) * math.pow(error_rate, filling_edit_distance) * math.comb(barcode_length, filling_edit_distance)
        density = barcode_count * math.pow(8 * barcode_length, filling_edit_distance) / math.pow(4, barcode_length)
        if error_probability > 0.01 or density > 1:
            filling_edit_distance -= 1

        if filling_edit_distance == 0:
            self.min_scores[base_name] = barcode_length - 1
        self.min_scores[base_name] = barcode_length - filling_edit_distance
        if variable_length:
            self.min_scores[base_name] = min(self.min_scores[base_name] + 1, barcode_length)
        logger.info("Minimal score for element %s is set to %d" % (base_name, self.min_scores[base_name]))

        if barcode_count > 1000000:
            logger.warning("The number of barcodes for element %s is large: %d, barcode calling may take substantial amount of time and RAM", (base_name, barcode_count))
            logger.warning("We suggest to use a sub-list of barcodes derived from short-read analysis whenever possible")

        kmer_size = find_optimal_kmer_size(barcode_length, barcode_count)
        if kmer_size <= 8:
            if barcode_count < 100000 or variable_length:
                self.index_dict[base_name] = ArrayKmerIndexer(barcode_list, kmer_size)
            else:
                self.index_dict[base_name] = Array2BitKmerIndexer(batch_str_to_2bit(barcode_list, barcode_length),
                                                                  kmer_size, seq_len=barcode_length)
        else:
            if barcode_count < 100000 or variable_length:
                self.index_dict[base_name] = KmerIndexer(barcode_list, kmer_size)
            else:
                self.index_dict[base_name] = Dict2BitKmerIndexer(batch_str_to_2bit(barcode_list, barcode_length),
                                                                 kmer_size, seq_len=barcode_length)
        logger.info("Indexed %d barcodes for element %s of length %d with k-mer size %d" % (barcode_count, base_name, barcode_length, kmer_size))

    def prepare_barcode_index_for_element(self, element: MoleculeElement):
        if element.element_type not in (ElementType.VAR_LIST, ElementType.VAR_FILE):
            return
        barcode_list = element.element_value
        self.prepare_barcode_index(element.element_name, barcode_list, element.element_length)

    def result_type(self):
        return ExtractionResult

    def header(self):
        """Get header matching ExtractionResult output format."""
        # Create a dummy result to get the correct header format
        dummy_result = ExtractionResult(self.molecule_structure, "", ".", {})
        return dummy_result.header()

    def find_barcode_umi(self, read_id, sequence):
        detected_elements_fwd = self._find_patterns_fwd(read_id, sequence)
        rev_seq = reverese_complement(sequence)
        detected_elements_rev = self._find_patterns_fwd(read_id, rev_seq)

        if self.has_polyt:
            fwd_has_polyt = ElementType.PolyT.name in detected_elements_fwd
            rev_has_polyt = ElementType.PolyT.name in detected_elements_rev
            if fwd_has_polyt and not rev_has_polyt:
                return ExtractionResult(self.molecule_structure, read_id, '+', detected_elements_fwd)
            if not fwd_has_polyt and rev_has_polyt:
                return ExtractionResult(self.molecule_structure, read_id, '-', detected_elements_rev)

        if len(detected_elements_fwd) >= len(detected_elements_rev):
            return ExtractionResult(self.molecule_structure, read_id, '+', detected_elements_fwd)
        return ExtractionResult(self.molecule_structure, read_id, '-', detected_elements_rev)

    def _find_patterns_fwd(self, read_id, sequence):
        logger.debug("== read id %s ==" % read_id)
        # TODO: make a list of DetectedElement rather that dict of str->DetectedElement for speed
        detected_elements = {}
        polyt_start, polyt_end = -1, -1
        if self.has_polyt:
            polyt_start, polyt_end = find_polyt(sequence)
            detected_elements[ElementType.PolyT.name] = DetectedElement(polyt_start, polyt_end, 0)
        logger.debug("PolyT %d" % polyt_start)

        self.detect_const_elements(sequence, polyt_start, polyt_end, detected_elements)
        concat_element_storage = {}
        dupl_element_storage = {}
        self.extract_variable_elements5(sequence, detected_elements, concat_element_storage, dupl_element_storage)
        self.extract_variable_elements3(sequence, detected_elements, concat_element_storage, dupl_element_storage)
        self.process_concatenated_elements(concat_element_storage, detected_elements, sequence)
        self.process_duplicated_elements(dupl_element_storage, detected_elements, sequence)
        logger.debug("== end read id %s ==" % read_id)
        return detected_elements

    def process_concatenated_elements(self, concatenated_element_storage, detected_results, sequence):
        for base_name, parts in concatenated_element_storage.items():
            # Check all parts were detected
            if any(start == -1 for start, end in parts):
                detected_results[base_name] = DetectedElement(-1, -1, -1, ExtractionResult.NOSEQ)
                continue

            # Concatenate sequences from all parts
            concatenated_seq = ""
            for start, end in parts:
                concatenated_seq += sequence[start:end + 1]

            # If correction is enabled and an index exists, correct against the whitelist
            if self.correct_sequences and base_name in self.index_dict:
                matching_sequences = self.index_dict[base_name].get_occurrences(concatenated_seq, max_hits=self.MAX_HITS)
                corrected_seq, seq_score, _, _ = find_candidate_with_max_score_ssw(
                    matching_sequences, concatenated_seq,
                    min_score=self.min_scores[base_name], score_diff=self.SCORE_DIFF)
                if corrected_seq is not None:
                    detected_results[base_name] = DetectedElement(
                        parts[0][0], parts[-1][1], seq_score, corrected_seq)
                else:
                    detected_results[base_name] = DetectedElement(-1, -1, -1, ExtractionResult.NOSEQ)
            else:
                # No correction: just store the raw concatenated sequence
                detected_results[base_name] = DetectedElement(
                    parts[0][0], parts[-1][1], 0, concatenated_seq)

    def process_duplicated_elements(self, duplicated_element_storage, detected_results, sequence):
        for base_name, parts in duplicated_element_storage.items():
            # Extract sequences for detected parts
            part_seqs = []
            part_coords = []
            for start, end in parts:
                if start != -1:
                    part_seqs.append(sequence[start:end + 1])
                    part_coords.append((start, end))
                else:
                    part_seqs.append(None)
                    part_coords.append((-1, -1))

            detected_seqs = [s for s in part_seqs if s is not None]
            if not detected_seqs:
                detected_results[base_name] = DetectedElement(-1, -1, -1, ExtractionResult.NOSEQ)
                continue

            # Use first detected copy's coordinates for reporting
            first_detected_idx = next(i for i, s in enumerate(part_seqs) if s is not None)

            if self.correct_sequences and base_name in self.index_dict:
                # Correct each copy independently
                corrected = []
                for seq in part_seqs:
                    if seq is None:
                        corrected.append((None, -1))
                        continue
                    matching = self.index_dict[base_name].get_occurrences(seq, max_hits=self.MAX_HITS)
                    corrected_seq, score, _, _ = find_candidate_with_max_score_ssw(
                        matching, seq, min_score=self.min_scores[base_name], score_diff=self.SCORE_DIFF)
                    corrected.append((corrected_seq, score))

                successful = [(seq, score) for seq, score in corrected if seq is not None]

                if not successful:
                    # Neither copy corrected
                    detected_results[base_name] = DetectedElement(-1, -1, -1, ExtractionResult.NOSEQ)
                elif len(successful) == 1:
                    # One succeeded
                    seq, score = successful[0]
                    detected_results[base_name] = DetectedElement(
                        part_coords[first_detected_idx][0],
                        part_coords[first_detected_idx][1],
                        score, seq)
                else:
                    # Multiple succeeded - check agreement
                    unique_seqs = set(seq for seq, _ in successful)
                    if len(unique_seqs) == 1:
                        # All agree
                        best_score = max(score for _, score in successful)
                        detected_results[base_name] = DetectedElement(
                            part_coords[first_detected_idx][0],
                            part_coords[first_detected_idx][1],
                            best_score, successful[0][0])
                    else:
                        # Copies disagree after correction
                        detected_results[base_name] = DetectedElement(-1, -1, -1, ExtractionResult.NOSEQ)
            else:
                # No correction: compare raw sequences
                if len(set(detected_seqs)) == 1:
                    # All match
                    detected_results[base_name] = DetectedElement(
                        part_coords[first_detected_idx][0],
                        part_coords[first_detected_idx][1],
                        0, detected_seqs[0])
                else:
                    # Differ: report all comma-separated
                    comma_sep = ",".join(detected_seqs)
                    detected_results[base_name] = DetectedElement(
                        part_coords[first_detected_idx][0],
                        part_coords[first_detected_idx][1],
                        0, comma_sep)

    def correct_element(self, element: MoleculeElement, sequence: str, potential_start: int, potential_end: int) -> DetectedElement:
        potential_seq = sequence[potential_start:potential_end + 1]
        matching_sequences = self.index_dict[element.element_name].get_occurrences(potential_seq, max_hits=self.MAX_HITS)
        corrected_seq, seq_score, seq_start, seq_end = \
            find_candidate_with_max_score_ssw(matching_sequences, potential_seq,
                                              min_score=self.min_scores[element.element_name], score_diff=self.SCORE_DIFF)
        if corrected_seq is not None:
            read_seq_start = potential_start + seq_start
            read_seq_end = potential_start + seq_end - 1
            return DetectedElement(read_seq_start, read_seq_end, seq_score, corrected_seq)
        else:
            return DetectedElement(-1, -1, -1, ExtractionResult.NOSEQ)

    def _backup_concatenated_element(self, element, potential_start, potential_end, element_storage):
        base_name, part_index = self.molecule_structure.elements_to_concatenate[element.element_name]
        if base_name not in element_storage:
            element_storage[base_name] = [(-1, -1) for _ in range(
                self.molecule_structure.concatenated_elements_counts[base_name])]
        # part_index is 1-based in MDF, convert to 0-based
        element_storage[base_name][part_index - 1] = (potential_start, potential_end)

    def _backup_duplicated_element(self, element, potential_start, potential_end, element_storage):
        base_name, part_index = self.molecule_structure.duplicated_elements[element.element_name]
        if base_name not in element_storage:
            element_storage[base_name] = [(-1, -1) for _ in range(
                self.molecule_structure.duplicated_elements_counts[base_name])]
        # part_index is 1-based in MDF, convert to 0-based
        element_storage[base_name][part_index - 1] = (potential_start, potential_end)

    def _detect_variable_element(self, el, sequence, potential_start, potential_end,
                                 detected_elements, concat_element_storage, dupl_element_storage,
                                 len_diff_threshold=None):
        potential_len = potential_end - potential_start + 1
        if len_diff_threshold is None or abs(potential_len - el.element_length) <= len_diff_threshold * el.element_length:
            if el.element_name in self.molecule_structure.elements_to_concatenate:
                self._backup_concatenated_element(el, potential_start, potential_end, concat_element_storage)
                return potential_start, potential_end

            if el.element_name in self.molecule_structure.duplicated_elements:
                self._backup_duplicated_element(el, potential_start, potential_end, dupl_element_storage)
                return potential_start, potential_end

            if el.element_name in self.elements_to_extract:
                detected_elements[el.element_name] = DetectedElement(potential_start, potential_end, 0,
                                                                     seq=sequence[potential_start:potential_end + 1])

            elif el.element_name in self.elements_to_correct:
                detected_element = self.correct_element(el, sequence, potential_start, potential_end)
                if detected_element.start != -1:
                    detected_elements[el.element_name] = detected_element
                    return detected_element.start, detected_element.end
        else:
            detected_elements[el.element_name] = DetectedElement(-1, -1, -1, ExtractionResult.NOSEQ)

        return potential_start, potential_end

    def _detect_potential_start(self, element, element_index, potential_end, detected_elements):
        if element_index == 0:
            potential_start = potential_end - element.element_length + 1
        else:
            prev_el = self.molecule_structure.ordered_elements[element_index - 1]
            if prev_el.element_name in detected_elements:
                potential_start = detected_elements[prev_el].end + 1
            else:
                potential_start = potential_end - element.element_length + 1
        if potential_start < 0:
            potential_start = 0

        return potential_start

    def _detect_potential_end(self, sequence, element, element_index, potential_start, detected_elements):
        if element_index + 1 == len(self.molecule_structure.ordered_elements):
            potential_end = potential_start + element.element_length - 1
        else:
            next_el = self.molecule_structure.ordered_elements[element_index + 1]
            if next_el.element_name in detected_elements:
                potential_end = detected_elements[next_el.element_name].start - 1
            else:
                potential_end = potential_start + element.element_length - 1
        if potential_end >= len(sequence):
            potential_end = len(sequence) - 1
        return potential_end

    def extract_variable_elements5(self, sequence, detected_elements, concat_element_storage, dupl_element_storage):
        first_detected_const_element = None
        for i, el in enumerate(self.molecule_structure):
            if el.element_type == ElementType.cDNA:
                break
            if (el.element_type == ElementType.PolyT or el.element_type.is_constant()) and el.element_name in detected_elements:
                first_detected_const_element = i
                break

        if first_detected_const_element is None: return

        # extracting elements preceding first detected const element
        current_pos = detected_elements[
                          self.molecule_structure.ordered_elements[first_detected_const_element].element_name].start - 1
        for i in range(first_detected_const_element - 1, -1, -1):
            el: MoleculeElement = self.molecule_structure.ordered_elements[i]

            if el.element_type.is_base_separator():
                current_pos -= el.element_length
                continue
            if not el.element_type.is_variable():
                # we do not skip undetected const elements on the open side (e.g. here we go to the left of the first detected const element)
                break

            potential_end = current_pos
            potential_start = potential_end - el.element_length + 1
            if potential_start < 0: break

            el_start, _ = self._detect_variable_element(el, sequence, potential_start, potential_end,
                                                        detected_elements, concat_element_storage, dupl_element_storage)
            current_pos = el_start - 1

        current_pos = detected_elements[
                          self.molecule_structure.ordered_elements[first_detected_const_element].element_name].end + 1
        for i in range(first_detected_const_element + 1, len(self.molecule_structure.ordered_elements)):
            if current_pos >= len(sequence):
                break

            el: MoleculeElement = self.molecule_structure.ordered_elements[i]
            if el.element_type in [ElementType.cDNA, ElementType.PolyT]:
                break
            elif el.element_type.is_base_separator():
                current_pos += el.element_length
                continue
            elif el.element_type.is_constant():
                if el.element_name in detected_elements:
                    current_pos = detected_elements[el.element_name].end + 1
                else:
                    current_pos += el.element_length
                continue

            potential_start = current_pos
            potential_end = self._detect_potential_end(sequence, el, i, potential_start, detected_elements)

            _, el_end = self._detect_variable_element(el, sequence, potential_start, potential_end,
                                                      detected_elements, concat_element_storage, dupl_element_storage,
                                                      len_diff_threshold=self.MAX_LEN_DIFF)
            current_pos = el_end + 1


    def extract_variable_elements3(self, sequence, detected_elements, concat_element_storage, dupl_element_storage):
        last_detected_const_element = None
        for i in range(len(self.molecule_structure.ordered_elements) - 1, -1, -1):
            el = self.molecule_structure.ordered_elements[i]
            if el.element_type == ElementType.cDNA:
                break
            if (el.element_type == ElementType.PolyT or el.element_type.is_constant()) and el.element_name in detected_elements:
                last_detected_const_element = i
                break

        if last_detected_const_element is None: return

        # extracting elements following the last detected const element
        current_pos = detected_elements[
                          self.molecule_structure.ordered_elements[last_detected_const_element].element_name].end + 1
        for i in range(last_detected_const_element + 1, len(self.molecule_structure.ordered_elements)):
            el: MoleculeElement = self.molecule_structure.ordered_elements[i]

            if el.element_type.is_base_separator():
                current_pos += el.element_length
                continue
            if not el.element_type.is_variable():
                # we do not skip undetected const elements on the open side (e.g. here we go to the left of the first detected const element)
                break

            potential_start = current_pos
            potential_end = potential_start + el.element_length - 1
            if potential_end >= len(sequence): break

            _, el_end = self._detect_variable_element(el, sequence, potential_start, potential_end,
                                                      detected_elements, concat_element_storage, dupl_element_storage)
            current_pos = el_end + 1

        current_pos = detected_elements[
                          self.molecule_structure.ordered_elements[last_detected_const_element].element_name].start - 1
        for i in range(last_detected_const_element - 1, -1, -1):
            if current_pos <= 0:
                break
            el: MoleculeElement = self.molecule_structure.ordered_elements[i]

            if el.element_type in [ElementType.cDNA, ElementType.PolyT]:
                break
            elif el.element_type.is_base_separator():
                current_pos -= el.element_length
                continue
            elif el.element_type.is_constant():
                if el.element_name in detected_elements:
                    current_pos = detected_elements[el.element_name].start - 1
                else:
                    current_pos -= el.element_length
                continue

            potential_end = current_pos
            potential_start = self._detect_potential_start(el, i, potential_end, detected_elements)

            el_start, _ = self._detect_variable_element(el, sequence, potential_start, potential_end,
                                                        detected_elements, concat_element_storage, dupl_element_storage,
                                                        len_diff_threshold=self.MAX_LEN_DIFF)
            current_pos = el_start - 1

    def detect_const_elements(self, sequence, polyt_start, polyt_end, detected_elements):
        # searching left of cDNA
        first_element_detected = False
        current_search_start = 0
        for el in self.molecule_structure:
            if el.element_type in [ElementType.cDNA, ElementType.PolyT]:
                break
            if el.element_name not in self.constant_elements_to_detect:
                continue

            search_start = current_search_start
            search_end = len(sequence) if polyt_start == -1 else polyt_start + 1
            element_occurrences = self.index_dict[el.element_name].get_occurrences_substr(sequence,
                                                                                          search_start,
                                                                                          search_end)
            if not first_element_detected:
                min_score = int(el.element_length * self.MIN_SCORE_COEFF_TERMMINAL)
                start_delta = -1
            else:
                min_score = int(el.element_length * self.MIN_SCORE_COEFF)
                start_delta = self.TERMINAL_MATCH_DELTA

            element_start, element_end = detect_exact_positions(sequence,
                                                                search_start, search_end,
                                                                self.index_dict[el.element_name].k,
                                                                el.element_value,
                                                                element_occurrences,
                                                                min_score=min_score,
                                                                start_delta=start_delta,
                                                                end_delta=self.TERMINAL_MATCH_DELTA)


            if element_start is not None:
                current_search_start = element_end
                if not first_element_detected:
                    first_element_detected = True

                detected_elements[el.element_name] = DetectedElement(element_start, element_end)

        last_element_detected = False
        current_search_end = len(sequence)
        for i in range(len(self.molecule_structure.ordered_elements) - 1, -1, -1):
            el = self.molecule_structure.ordered_elements[i]
            if el.element_type in [ElementType.cDNA, ElementType.PolyT]:
                break
            if el.element_name not in self.constant_elements_to_detect:
                continue

            search_start = 0 if polyt_start == -1 else polyt_end + 1
            search_end = current_search_end
            element_occurrences = self.index_dict[el.element_name].get_occurrences_substr(sequence, search_start,
                                                                                          search_end)

            if not last_element_detected:
                min_score = int(el.element_length * self.MIN_SCORE_COEFF_TERMMINAL)
                end_delta = -1
            else:
                min_score = int(el.element_length * self.MIN_SCORE_COEFF)
                end_delta = self.TERMINAL_MATCH_DELTA

            element_start, element_end = detect_exact_positions(sequence,
                                                                search_start, search_end,
                                                                self.index_dict[el.element_name].k,
                                                                el.element_value,
                                                                element_occurrences,
                                                                min_score=min_score,
                                                                start_delta=self.TERMINAL_MATCH_DELTA,
                                                                end_delta=end_delta)
            if element_start is not None:
                current_search_end = element_end
                if not last_element_detected:
                    last_element_detected = True

                detected_elements[el.element_name] = DetectedElement(element_start, element_end)



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import gzip

from .common import (
    CANONICAL_FWD_SITES,
    CANONICAL_REV_SITES,
    argmin,
    junctions_from_blocks,
    range_list_to_str,
    sum_intervals_from_point,
    sum_intervals_to_point
)
from .serialization import (
    write_short_int,
    read_short_int,
    SHORT_TERMINATION_INT
)
from .isoform_assignment import (match_subtype_to_str_with_additional_info,
                                 ReadAssignment,
                                 BasicReadAssignment,
                                 MatchClassification,
                                 ReadAssignmentType)
from .gene_info import GeneInfo, GeneList


logger = logging.getLogger('IsoQuant')


class PrintAllFunctor:
    def check(self, assignment):
        return True


class PrintOnlyFunctor:
    def __init__(self, allowed_types):
        if isinstance(allowed_types, list):
            self.allowed_types = set(allowed_types)
        elif isinstance(allowed_types, set):
            self.allowed_types = allowed_types
        else:
            self.allowed_types = {allowed_types}

    def check(self, assignment):
        if assignment is None:
            return False
        return assignment.assignment_type in self.allowed_types


class AbstractAssignmentPrinter:
    def __init__(self, output_file_name, params, assignment_checker=PrintAllFunctor()):
        self.assignment_checker = assignment_checker
        self.params = params
        self.output_file_name = output_file_name

    def add_read_info(self, read_assignment):
        raise NotImplementedError()

    def flush(self):
        raise NotImplementedError()


class TextFileAssignmentPrinter(AbstractAssignmentPrinter):
    def __init__(self, output_file_name, params, assignment_checker=PrintAllFunctor(), gzipped=False):
        AbstractAssignmentPrinter.__init__(self, output_file_name, params, assignment_checker)
        self.gzipped = gzipped
        if gzipped:
            self.output_file_name += ".gz"
            self.output_file = gzip.open(self.output_file_name, "wt")
        else:
            self.output_file = open(self.output_file_name, "w")

    def __del__(self):
        self.output_file.close()

    def flush(self):
        if not self.gzipped:
            self.output_file.flush()


class ReadAssignmentCompositePrinter:
    def __init__(self, printers):
        self.printers = printers

    def add_read_info(self, read_assignment):
        for p in self.printers:
            p.add_read_info(read_assignment)

    def flush(self):
        for p in self.printers:
            p.flush()


class VoidPrinter:
    def add_read_info(self, _):
        pass

    def flush(self):
        pass


# write mapped reads to bed file
class BEDPrinter(TextFileAssignmentPrinter):
    def __init__(self, output_file_name, params, print_corrected=False, assignment_checker=PrintAllFunctor(),
                 gzipped=False):
        TextFileAssignmentPrinter.__init__(self, output_file_name, params, assignment_checker, gzipped=gzipped)
        self.print_corrected = print_corrected
        self.output_file.write("#chrom\tchromStart\tchromEnd\tname\tscore\tstrand\tthickStart\tthickEnd\titemRgb\tblockCount\tblockSizes\tblockStarts\n")

    def add_read_info(self, read_assignment):
        if read_assignment is None or read_assignment.assignment_type is None or \
                not hasattr(read_assignment, "gene_info") or read_assignment.gene_info is None:
            return
        if self.assignment_checker is None or not self.assignment_checker.check(read_assignment):
            return

        strand = read_assignment.mapped_strand
        chr_id = read_assignment.gene_info.chr_id
        exon_blocks = read_assignment.corrected_exons if self.print_corrected else read_assignment.exons

        self.output_file.write("%s\t%d\t%d\t%s\t0\t%s\t%d\t%d\t%d\t%d\t%s\t%s\n" %
                           (chr_id, exon_blocks[0][0] - 1, exon_blocks[-1][1],
                            read_assignment.read_id, strand,
                            exon_blocks[0][0] - 1, exon_blocks[0][0] - 1, 0,
                            len(exon_blocks),
                            ",".join([str(e[1] - e[0] + 1) for e in exon_blocks]),
                            ",".join([str(e[0] - exon_blocks[0][0]) for e in exon_blocks])))


class TmpFileAssignmentPrinter(AbstractAssignmentPrinter):
    GENE_INFO = 255
    READ_ASSIGNMENT = 255 << 8

    def __init__(self, output_file_name, params):
        AbstractAssignmentPrinter.__init__(self, output_file_name, params)
        self.dumper = open(self.output_file_name, "wb")

    def __del__(self):
        self.close()

    def close(self):
        if not self.dumper.closed:
            write_short_int(SHORT_TERMINATION_INT, self.dumper)
            self.dumper.close()

    def add_gene_info(self, gene_info):
        write_short_int(self.GENE_INFO, self.dumper)
        gene_info.serialize(self.dumper)

    def add_read_info(self, read_assignment):
        if read_assignment.assignment_type is None or read_assignment.isoform_matches is None:
            return
        write_short_int(self.READ_ASSIGNMENT, self.dumper)
        read_assignment.serialize(self.dumper)

    def flush(self):
        pass


class BaseTmpFileAssignmentLoader:
    def __init__(self, input_file_name):
        self.loader = open(input_file_name, "rb")
        self.current_id = None
        self._read_id()

    def __del__(self):
        self.loader.close()

    def _read_id(self):
        self.current_id = read_short_int(self.loader)

    def has_next(self):
        return self.current_id != SHORT_TERMINATION_INT

    def is_gene_info(self):
        return self.current_id == TmpFileAssignmentPrinter.GENE_INFO

    def is_read_assignment(self):
        return self.current_id == TmpFileAssignmentPrinter.READ_ASSIGNMENT


class NormalTmpFileAssignmentLoader(BaseTmpFileAssignmentLoader):
    def __init__(self, input_file_name, genedb, chr_record, string_pools=None):
        BaseTmpFileAssignmentLoader.__init__(self, input_file_name)
        self.genedb = genedb
        self.chr_record = chr_record
        self.current_gene_info = None
        self.string_pools = string_pools

    def get_object(self):
        if self.is_gene_info():
            self.current_gene_info = GeneInfo.deserialize(self.loader, self.genedb)
            if self.chr_record:
                self.current_gene_info.set_reference_sequence(self.current_gene_info.all_read_region_start,
                                                              self.current_gene_info.all_read_region_end,
                                                              self.chr_record)
            self._read_id()
            return self.current_gene_info
        elif self.is_read_assignment():
            assert self.current_gene_info is not None
            assignment = ReadAssignment.deserialize(self.loader, self.current_gene_info, self.string_pools)
            self._read_id()
            return assignment
        else:
            return None


class GeneListTmpFileAssignmentLoader(BaseTmpFileAssignmentLoader):
    def __init__(self, input_file_name, string_pools=None):
        BaseTmpFileAssignmentLoader.__init__(self, input_file_name)
        self.current_gene_info = None
        self.string_pools = string_pools

    def get_object(self):
        if self.is_gene_info():
            self.current_gene_info = GeneList.deserialize(self.loader)
            self._read_id()
            return self.current_gene_info
        elif self.is_read_assignment():
            assert self.current_gene_info is not None
            assignment = ReadAssignment.deserialize(self.loader, self.current_gene_info, self.string_pools)
            self._read_id()
            return assignment
        else:
            return None


class QuickTmpFileAssignmentLoader(BaseTmpFileAssignmentLoader):
    def __init__(self, input_file_name, string_pools=None):
        BaseTmpFileAssignmentLoader.__init__(self, input_file_name)
        self.string_pools = string_pools

    def get_object(self):
        if self.is_gene_info():
            GeneInfo.deserialize(self.loader, None)
            self._read_id()
            return None
        elif self.is_read_assignment():
            assignment = BasicReadAssignment.deserialize_from_read_assignment(self.loader, self.string_pools)
            self._read_id()
            return assignment
        else:
            return None


class BasicTSVAssignmentPrinter(TextFileAssignmentPrinter):
    def __init__(self, output_file_name, params, io_support, additional_header = "", gzipped=False):
        TextFileAssignmentPrinter.__init__(self, output_file_name, params, gzipped=gzipped)
        self.header = "read_id\tchr\tstrand\tisoform_id\tgene_id" \
                      "\tassignment_type\tassignment_events\texons\tadditional_info\tgroups\n"
        self.output_file.write(additional_header)
        self.output_file.write(self.header)
        self.io_support = io_support

    @staticmethod
    def unmatched_line(read_assignment, additional_info):
        read_exons = read_assignment.exons
        strand = read_assignment.strand
        line = read_assignment.read_id + "\t" + read_assignment.chr_id + "\t" + strand + "\t.\t.\t" + \
               read_assignment.assignment_type.name + "\t.\t" + range_list_to_str(read_exons)

        for attr in read_assignment.additional_attributes.keys():
            val = read_assignment.additional_attributes[attr]
            additional_info.append("%s=%s;" % (attr, val))
        if additional_info:
            line += "\t" + " ".join(additional_info)
        else:
            line += "\t*"
        line += "\t%s\n" % ",".join(read_assignment.read_group)
        return line

    def add_read_info(self, read_assignment):
        if read_assignment is None:
            return
        if self.assignment_checker is None or not self.assignment_checker.check(read_assignment):
            return

        if read_assignment.assignment_type is None:
            logger.warning("Empty assignment read id %s" %  (read_assignment.read_id))
            return
        elif read_assignment.exons is None:
            logger.warning("Empty combined profile, read id %s, assignment type %s" %
                           (read_assignment.read_id, str(read_assignment.assignment_type)))
            return

        read_exons = read_assignment.exons
        read_introns = junctions_from_blocks(read_exons)

        if not read_assignment.isoform_matches:
            self.output_file.write(BasicTSVAssignmentPrinter.unmatched_line(read_assignment, []))
            return

        for m in read_assignment.isoform_matches:
            if m.assigned_transcript is None:
                self.output_file.write(
                    BasicTSVAssignmentPrinter.unmatched_line(
                        read_assignment, ["Classification=" + str(m.match_classification.name) + ";"]))
                continue
            isoform_introns = read_assignment.gene_info.all_isoforms_introns[m.assigned_transcript]
            event_string = ",".join([match_subtype_to_str_with_additional_info(x, read_assignment.strand,
                                                                               read_introns, isoform_introns)
                                     for x in m.match_subclassifications])
            strand = read_assignment.strand
            line = (read_assignment.read_id + "\t" + read_assignment.chr_id + "\t" + strand + "\t" +
                    m.assigned_transcript + "\t" + m.assigned_gene + "\t" +
                    read_assignment.assignment_type.name + "\t" + event_string + "\t" +
                    range_list_to_str(read_exons))

            additional_info = []
            additional_info.append("gene_assignment=" + str(read_assignment.gene_assignment_type.name) + ";")
            additional_info.append("PolyA=" + str(read_assignment.polyA_found) + ";")
            if self.params.cage is not None:
                additional_info.append("CAGE=" + str(read_assignment.cage_found) + ";")
            if self.params.check_canonical and read_assignment.gene_info.reference_region:
                if len(read_introns) == 0:
                    additional_info.append("Canonical=Unspliced;")
                else:
                    all_canonical = self.io_support.check_sites_are_canonical(read_introns, read_assignment.gene_info, strand)
                    additional_info.append("Canonical=" + str(all_canonical) + ";")

            additional_info.append("Classification=" + str(m.match_classification.name) + ";")

            for attr in read_assignment.additional_attributes.keys():
                val = read_assignment.additional_attributes[attr]
                additional_info.append("%s=%s;" % (attr, val))

            if additional_info:
                line += "\t" + " ".join(additional_info)
            else:
                line += "\t*"
            line += "\t%s\n" % ",".join(read_assignment.read_group)
            self.output_file.write(line)


    """ sqanti output
    
    isoform: the isoform ID. Usually in PB.X.Y format.
    chrom: chromosome.
    strand: strand.
    length: isoform length.
    exons: number of exons.
    structural_category: one of the categories ["full-splice_match", "incomplete-splice_match", "novel_in_catalog", "novel_not_in_catalog", "genic", "antisense", "fusion", "intergenic", "genic_intron"]
    associated_gene: the reference gene name.
    associated_transcript: the reference transcript name.
    ref_length: reference transcript length.
    ref_exons: reference transcript number of exons.
    diff_to_TSS: distance of query isoform 5' start to reference transcript start end. Negative value means query starts downstream of reference.
    diff_to_TTS: distance of query isoform 3' end to reference annotated end site. Negative value means query ends upstream of reference.
    diff_to_gene_TSS: distance of query isoform 5' start to the closest start end of any transcripts of the matching gene. This field is different from diff_to_TSS since it's looking at all annotated starts of a gene. Negative value means query starts downstream of reference.
    diff_to_gene_TTS: distance of query isoform 3' end to the closest end of any transcripts of the matching gene. Negative value means query ends upstream of reference.
    subcategory: additional splicing categorization, separated by semi-colons. Categories include: mono-exon, multi-exon. Intron rentention is marked with intron_retention.
    RTS_stage: TRUE if one of the junctions could be a RT switching artifact.
    all_canonical: TRUE if all junctions have canonical splice sites.
    - min_sample_cov: sample with minimum coverage.
    - min_cov: minimum junction coverage based on short read STAR junction output file. NA if no short read given.
    - min_cov_pos: the junction that had the fewest coverage. NA if no short read data given.
    - sd_cov: standard deviation of junction coverage counts from short read data. NA if no short read data given.
    - FL or FL.<sample>: FL count associated with this isoform per sample if --fl_count is provided, otherwise NA.
    n_indels: total number of indels based on alignment.
    n_indels_junc: number of junctions in this isoform that have alignment indels near the junction site (indicating potentially unreliable junctions).
    bite: TRUE if all junctions match reference junctions completely.
    - iso_exp: short read expression for this isoform if --expression is provided, otherwise NA.
    - gene_exp: short read expression for the gene associated with this isoform (summing over all isoforms) if --expression is provided, otherwise NA.
    - ratio_exp: ratio of iso_exp to gene_exp if --expression is provided, otherwise NA.
    - FSM_class: ignore this field for now.
    - ORF_length: predicted ORF length.
    - CDS_length: predicted CDS length.
    - CDS_start: CDS start.
    - CDS_end: CDS end.
    CDS_genomic_start: genomic coordinate of the CDS start. If on - strand, this coord will be greater than the end.
    CDS_genomic_end: genomic coordinate of the CDS end. If on - strand, this coord will be smaller than the start.
    - predicted_NMD: TRUE if there's a predicted ORF and CDS ends before the last junction; FALSE if otherwise. NA if non-coding.
    perc_A_downstreamTTS: percent of genomic "A"s in the downstream 20 bp window. If this number if high (say > 0.8), the 3' end site of this isoform is probably not reliable.
    seq_A_downstream_TTS: sequence of the downstream 20 bp window.
    dist_to_cage_peak: distance to closest TSS based on CAGE Peak data. Negative means upstream of TSS and positive means downstream of TSS. Strand-specific. SQANTI2 only searches for nearby CAGE Peaks within 10000 bp of the PacBio transcript start site. Will be NA if none are found within 10000 bp.
    within_cage_peak: TRUE if the PacBio transcript start site is within a CAGE Peak.
    dist_to_polya_site: distance to closest polyA site based on PolyA site data. Negative means upstream of closest site and positive means downstream of closest site. Strand-specific. SQANTI2 only searches for nearby CAGE Peaks within 100 bp of the PacBio transcript start site. Will be NA if none are found within 100 bp.
    within_polya_site: TRUE if the PacBio transcript polyA site is within 100 bp (upstream or downstream) of an annotated polyA site.
    polyA_motif: if --polyA_motif_list is given, shows the top ranking polyA motif found within 50 bp upstream of end.
    polyA_dist: if --polyA_motif_list is given, shows the location of the last base of the hexamer. Position 0 is the putative poly(A) site. This distance is hence always negative because it is upstream.
    
    """


class SqantiTSVPrinter(TextFileAssignmentPrinter):
    def __init__(self, output_file_name, params, io_support):
        TextFileAssignmentPrinter.__init__(self, output_file_name, params)
        self.header = 'isoform\tchrom\tstrand\tlength\texons\tstructural_category' \
                      '\tassociated_gene\tassociated_transcript\tref_length\tref_exons\tdiff_to_TSS\tdiff_to_TTS' \
                      '\tdiff_to_gene_TSS\tdiff_to_gene_TTS\tsubcategory\tRTS_stage\tall_canonical' \
                      '\tmin_sample_cov\tmin_cov\tmin_cov_pos\tsd_cov\tFL' \
                      '\tn_indels\tn_indels_junc\tbite\tiso_exp\tgene_exp\tratio_exp\tFSM_class' \
                      '\tcoding\tORF_length\tCDS_length\tCDS_start\tCDS_end' \
                      '\tCDS_genomic_start\tCDS_genomic_end\tpredicted_NMD' \
                      '\tperc_A_downstream_TTS\tseq_A_downstream_TTS\tdist_to_CAGE_peak' \
                      '\twithin_CAGE_peak\tdist_to_polyA_site\twithin_polyA_site\tpolyA_motif\tpolyA_dist' \
                      '\tpolyA_motif_found\tORF_seq\tratio_TSS\n'
        self.output_file.write(self.header)
        self.output_file.flush()
        self.io_support = io_support

    def add_read_info(self, read_assignment):
        if read_assignment is None:
            return

        if read_assignment.assignment_type == ReadAssignmentType.noninformative:
            return

        gene_info = read_assignment.gene_info
        if not read_assignment.isoform_matches:
            assigned = False
            gene_id = "NA"
            transcript_id = "NA"
            sqanti_classification = MatchClassification.intergenic
        else:
            assigned = True
            match = read_assignment.isoform_matches[0]
            gene_id = match.assigned_gene
            transcript_id = match.assigned_transcript
            sqanti_classification = match.match_classification
            if transcript_id is None:
                transcript_id = "NA"
                gene_id = "NA"
                assigned = False
                sqanti_classification = MatchClassification.intergenic
        strand = read_assignment.strand

        # FIXME not genomic distance
        if assigned:
            dist_to_tss = self.io_support.count_tss_dist(read_assignment, transcript_id)
            dist_to_tts = self.io_support.count_tts_dist(read_assignment, transcript_id)
            dist_to_gene_tss, dist_to_gene_tts = self.io_support.find_closests_tsts(read_assignment)
        else:
            dist_to_tss, dist_to_tts = "NA", "NA"
            dist_to_gene_tss, dist_to_gene_tts = "NA", "NA"
        if strand == '-':
            # swapping distances since 5' and 3' are now opposite
            dist_to_tss, dist_to_tts = dist_to_tts, dist_to_tss
            dist_to_gene_tss, dist_to_gene_tts = dist_to_gene_tts, dist_to_gene_tss

        if assigned:
            subtypes = ";".join([x.event_type.name for x in match.match_subclassifications])
            if not subtypes: subtypes = "."
        else:
            subtypes = "."

        indel_count = read_assignment.additional_info["indel_count"] \
            if "indel_count" in read_assignment.additional_info.keys() else "-1"
        junctions_with_indels = read_assignment.additional_info["junctions_with_indels"] \
            if "junctions_with_indels" in read_assignment.additional_info.keys() else "-1"
        bite = self.io_support.check_all_sites_match_reference(read_assignment)
        if assigned:
            ref_cds_start, ref_cds_end = self.io_support.find_ref_CDS_region(gene_info, transcript_id)
        else:
            ref_cds_start, ref_cds_end = "NA", "NA"

        if gene_info.reference_region:
            read_exons = read_assignment.exons
            read_introns = junctions_from_blocks(read_exons)
            all_canonical = str(self.io_support.check_sites_are_canonical(read_introns, gene_info, strand))
            seq_A_downstream_TTS, perc_A_downstreamTTS = \
                self.io_support.check_downstream_polya((read_assignment.start(), read_assignment.end()), gene_info, strand)
            perc_A_downstreamTTS = "%0.2f" % perc_A_downstreamTTS
        else:
            all_canonical = "NA"
            perc_A_downstreamTTS = "NA"
            seq_A_downstream_TTS = "NA"

        # TODO
        RTS = "FALSE"
        min_sample_cov  = "NA"
        min_cov = "NA"
        min_cov_pos = "NA"
        sd_cov = "NA"
        FL = "NA"
        iso_exp = "NA"
        gene_exp = "NA"
        ratio_exp = "NA"
        FSM_class = read_assignment.additional_info["FSM_class"]
        coding = "NA"
        ORF_length = "NA"
        CDS_length = "NA"
        CDS_start = "NA"
        CDS_end = "NA"
        predicted_NMD = "NA"
        dist_to_cage_peak = "NA"
        within_cage_peak = "NA"
        dist_to_polya_site = "NA"
        within_polya_site = "NA"
        polyA_motif = "NA"
        polyA_dist = "NA"
        polyA_motif_found = "NA"
        ORF_seq = "NA"
        ratio_TSS = "NA"

        transcript_len = "NA" if not assigned else gene_info.total_transcript_length(transcript_id)
        transcript_exon_count = "NA" if not assigned else gene_info.transcript_exon_count(transcript_id)

        value_list = [read_assignment.read_id, gene_info.chr_id, strand, read_assignment.length(),
                      read_assignment.exon_count(), sqanti_classification.name, gene_id, transcript_id,
                      transcript_len, transcript_exon_count,
                      dist_to_tss, dist_to_tts, dist_to_gene_tss, dist_to_gene_tts, subtypes, RTS,
                      all_canonical, min_sample_cov, min_cov, min_cov_pos, sd_cov, FL,
                      indel_count, junctions_with_indels, bite, iso_exp, gene_exp, ratio_exp, FSM_class,
                      coding, ORF_length, CDS_length, CDS_start, CDS_end,
                      ref_cds_start, ref_cds_end, predicted_NMD, perc_A_downstreamTTS, seq_A_downstream_TTS,
                      dist_to_cage_peak, within_cage_peak, dist_to_polya_site,
                      within_polya_site, polyA_motif, polyA_dist, polyA_motif_found, ORF_seq, ratio_TSS]
        l = "\t".join([str(x) for x in value_list])
        self.output_file.write(l + "\n")

    def __del__(self):
        self.output_file.close()

    def flush(self):
        self.output_file.flush()


class IOSupport:
    def __init__(self, params):
        self.params = params

    def count_tss_dist(self, read_assignment, transcript_id):
        transcript_start = read_assignment.gene_info.transcript_start(transcript_id)
        read_start = read_assignment.start()
        if read_start < transcript_start:
            return -sum_intervals_to_point(read_assignment.exons, transcript_start)
        else:
            return sum_intervals_to_point(read_assignment.gene_info.all_isoforms_exons[transcript_id], read_start)

    def count_tts_dist(self, read_assignment, transcript_id):
        transcript_end = read_assignment.gene_info.transcript_end(transcript_id)
        read_end = read_assignment.end()
        if read_end > transcript_end:
            return -sum_intervals_from_point(read_assignment.exons, transcript_end)
        else:
            return sum_intervals_from_point(read_assignment.gene_info.all_isoforms_exons[transcript_id], read_end)

    def find_closests_tsts(self, read_assignment):
        gene_info = read_assignment.gene_info
        match = read_assignment.isoform_matches[0]
        gene_id = match.assigned_gene

        dists_to_gene_tss = []
        dists_to_gene_tts = []
        for t in gene_info.db.children(gene_info.db[gene_id], featuretype=('transcript', 'mRNA'), order_by='start'):
            dists_to_gene_tss.append(self.count_tss_dist(read_assignment, t.id))
            dists_to_gene_tts.append(self.count_tts_dist(read_assignment, t.id))
        dist_to_gene_tss = dists_to_gene_tss[argmin([abs(x) for x in dists_to_gene_tss])]
        dist_to_gene_tts = dists_to_gene_tts[argmin([abs(x) for x in dists_to_gene_tts])]
        return dist_to_gene_tss, dist_to_gene_tts

    def find_ref_CDS_region(self, gene_info, transcript_id):
        cds = [c for c in gene_info.db.children(gene_info.db[transcript_id], featuretype='CDS', order_by='start')]
        if len(cds) == 0:
            return -1, -1
        if gene_info.isoform_strands[transcript_id] == "-":
            return cds[-1].end, cds[0].start
        return cds[0].start, cds[-1].end

    def check_all_sites_match_reference(self, read_assignment):
        # FIXME: make a new profile with delta = 0
        if len(read_assignment.exons) == 1:
            return "Unspliced"
        introns_match = read_assignment.introns_match
        return str(introns_match)

    # TODO: use built-in function for extracting ref seq
    def check_sites_are_canonical(self, read_introns, gene_info, strand):
        for intron in read_introns:
            if intron not in gene_info.canonical_sites:
                intron_left_pos = intron[0] - gene_info.all_read_region_start
                intron_right_pos = intron[1] - gene_info.all_read_region_start
                left_site = gene_info.reference_region[intron_left_pos:intron_left_pos+2]
                right_site = gene_info.reference_region[intron_right_pos - 1:intron_right_pos + 1]
                if strand == '+':
                    gene_info.canonical_sites[intron] = (left_site, right_site) in CANONICAL_FWD_SITES
                else:
                    gene_info.canonical_sites[intron] = (left_site, right_site) in CANONICAL_REV_SITES

            if not gene_info.canonical_sites[intron]:
                return False
        return True

    def add_canonical_info_for_model(self, model, gene_info):
        if not gene_info.reference_region:
            return
        key_word = 'Canonical'
        if model.check_additional(key_word):
            return
        model_introns = junctions_from_blocks(model.exon_blocks)
        if len(model_introns) == 0:
            canonical_info = "Unspliced"
        else:
            all_canonical = self.check_sites_are_canonical(model_introns, gene_info, model.strand)
            canonical_info = str(all_canonical)
        model.add_additional_attribute(key_word, canonical_info)

    def add_canonical_info(self, model_storage, gene_info):
        for m in model_storage:
            self.add_canonical_info_for_model(m, gene_info)

    def check_downstream_polya(self, read_coords, gene_info, strand):
        if strand == '+':
            read_end = read_coords[1] - gene_info.all_read_region_start + 1
            seq = gene_info.reference_region[read_end:read_end + self.params.upstream_region_len]
            a_percentage = float(seq.upper().count('A')) / float(self.params.upstream_region_len)
        else:
            read_start = read_coords[0] - gene_info.all_read_region_start
            seq = gene_info.reference_region[read_start - self.params.upstream_region_len:read_start]
            a_percentage = float(seq.upper().count('T')) / float(self.params.upstream_region_len)
        return seq, a_percentage



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Memory-efficient k-mer indexers using 2-bit DNA encoding.

Dict2BitKmerIndexer: Dictionary with integer keys (~40% less memory than KmerIndexer)
Array2BitKmerIndexer: Flat array with range indexing for large barcode sets
"""

import math
from typing import List, Tuple, Iterable, Dict, DefaultDict
from collections import defaultdict
from ..common import bit_to_str, str_to_2bit


class Dict2BitKmerIndexer:
    """
    Memory-efficient k-mer indexer using 2-bit encoded sequences and integer keys.

    Stores sequences as 2-bit encoded integers (8 bytes per 25bp barcode vs 25+ bytes for strings).
    Uses dictionary for k-mer index (vs 4^k array in Array2BitKmerIndexer).
    Best for medium to large barcode sets.
    """

    def __init__(self, known_bit_seqs: Iterable[int], kmer_size: int, seq_len: int):
        """
        Initialize k-mer index with 2-bit encoded sequences.

        Args:
            known_bit_seqs: Collection of 2-bit encoded sequences (use str_to_2bit or batch_str_to_2bit)
            kmer_size: Length of k-mers to use for indexing
            seq_len: Length of sequences in nucleotides
        """
        self.seq_list: List[int] = list(known_bit_seqs)
        self.k: int = kmer_size
        self.seq_len: int = seq_len
        self.mask: int = (1 << (2 * self.k)) - 1
        self.index: DefaultDict[int, List[int]] = defaultdict(list)
        self._index()

    def _get_kmer_indexes_from_bits(self, bin_seq: int) -> Iterable[int]:
        """Extract k-mer indices from a 2-bit encoded sequence."""
        # Convert numpy scalar to Python int for bitwise operations
        bin_seq = int(bin_seq)
        for i in range(self.seq_len - self.k + 1):
            yield (bin_seq >> ((self.seq_len - self.k - i) * 2)) & self.mask

    def _get_kmer_indexes_from_str(self, seq: str) -> Iterable[int]:
        """Generate binary-encoded k-mer indices from string using sliding window."""
        if len(seq) < self.k:
            return
        # Use same encoding as common.py: (ord(char) & 6) >> 1
        # A=0, C=1, T=2, G=3
        # Initialize first k-mer
        kmer_idx = 0
        for i in range(self.k):
            kmer_idx |= ((ord(seq[i]) & 6) >> 1) << ((self.k - i - 1) * 2)
        yield kmer_idx
        # Slide window
        for i in range(self.k, len(seq)):
            kmer_idx = ((kmer_idx << 2) & self.mask) | ((ord(seq[i]) & 6) >> 1)
            yield kmer_idx

    def _index(self) -> None:
        """Build k-mer index from all sequences."""
        for i, bin_seq in enumerate(self.seq_list):
            for kmer_idx in self._get_kmer_indexes_from_bits(bin_seq):
                self.index[kmer_idx].append(i)

    def empty(self) -> bool:
        """Check if index is empty."""
        return len(self.seq_list) == 0

    def get_occurrences(self, sequence: str, max_hits: int = 0, min_kmers: int = 1,
                       hits_delta: int = 1, ignore_equal: bool = False) -> List[Tuple[str, int, List[int]]]:
        """
        Find indexed sequences with shared k-mers.

        Args:
            sequence: Query sequence (string) to search
            max_hits: Maximum number of results (0 = unlimited)
            min_kmers: Minimum shared k-mers required
            hits_delta: Include results within this many k-mers of top hit
            ignore_equal: Skip exact matches

        Returns:
            List of (sequence_str, shared_kmer_count, kmer_positions) tuples
        """
        barcode_counts: DefaultDict[int, int] = defaultdict(int)
        barcode_positions: DefaultDict[int, List[int]] = defaultdict(list)

        query_bin = str_to_2bit(sequence)
        for pos, kmer_idx in enumerate(self._get_kmer_indexes_from_str(sequence)):
            for seq_index in self.index.get(kmer_idx, []):
                barcode_counts[seq_index] += 1
                barcode_positions[seq_index].append(pos)

        result = []
        for seq_index in barcode_counts.keys():
            count = barcode_counts[seq_index]
            if count < min_kmers:
                continue
            bin_seq = self.seq_list[seq_index]
            if ignore_equal and bin_seq == query_bin:
                continue
            result.append((bin_seq, count, barcode_positions[seq_index]))

        if not result:
            return []

        top_hits = max(result, key=lambda x: x[1])[1]
        result = filter(lambda x: x[1] >= top_hits - hits_delta, result)
        result = sorted(result, reverse=True, key=lambda x: x[1])

        # Convert 2-bit sequences back to strings
        if max_hits == 0:
            return [(bit_to_str(x[0], self.seq_len), x[1], x[2]) for x in result]
        return [(bit_to_str(x[0], self.seq_len), x[1], x[2]) for x in list(result)[:max_hits]]


class Array2BitKmerIndexer:
    """
    Memory-efficient k-mer indexer using 2-bit encoding for both k-mers and sequences.

    Stores sequences as integers (2 bits per nucleotide) to minimize memory.
    Uses flat array with range indexing for better cache performance.
    Best for large barcode sets (e.g., single-cell whitelists).
    """

    def __init__(self, known_bin_seq: Iterable[int], kmer_size: int, seq_len: int):
        """
        Initialize 2-bit k-mer index.

        Args:
            known_bin_seq: Pre-encoded sequences as integers (use str_to_2bit)
            kmer_size: Length of k-mers
            seq_len: Length of sequences (all must be same length)
        """
        self.k: int = kmer_size
        total_kmers = int(math.pow(4, kmer_size))
        tmp_index: List[List[int]] = [[] for _ in range(total_kmers)]
        self.mask: int = (1 << (2 * self.k)) - 1
        self.seq_len: int = seq_len
        self.seq_mask: int = (1 << (2 * self.seq_len)) - 1
        self.total_sequences: int = 0
        self._index(known_bin_seq, tmp_index)

        # Flatten index for better cache performance
        self.index: List[int] = []
        self.index_ranges: List[int] = [0]
        for l in tmp_index:
            self.index += l
            self.index_ranges.append(len(self.index))

    def _get_kmer_bin_indexes(self, bin_seq: int) -> Iterable[int]:
        """Extract k-mer indices from a 2-bit encoded sequence."""
        for i in range(self.seq_len - self.k + 1):
            yield (bin_seq >> ((self.seq_len - self.k - i) * 2)) & self.mask

    def _index(self, known_bin_seq: Iterable[int], tmp_index: List[List[int]]) -> None:
        """Build k-mer index from 2-bit encoded sequences."""
        for bin_seq in known_bin_seq:
            self.total_sequences += 1
            for kmer_idx in self._get_kmer_bin_indexes(bin_seq):
                tmp_index[kmer_idx].append(bin_seq)

    def empty(self) -> bool:
        """Check if index is empty."""
        return self.total_sequences == 0

    def get_occurrences(self, sequence: str, max_hits: int = 0, min_kmers: int = 1,
                       hits_delta: int = 1, ignore_equal: bool = False) -> List[Tuple[str, int, List[int]]]:
        """
        Find indexed sequences with shared k-mers.

        Args:
            sequence: Query sequence (string, will be converted to 2-bit)
            max_hits: Maximum number of results (0 = unlimited)
            min_kmers: Minimum shared k-mers required
            hits_delta: Include results within this many k-mers of top hit
            ignore_equal: Skip exact matches

        Returns:
            List of (sequence_str, shared_kmer_count, kmer_positions) tuples
        """
        barcode_counts: DefaultDict[int, int] = defaultdict(int)
        barcode_positions: DefaultDict[int, List[int]] = defaultdict(list)

        seq = str_to_2bit(sequence)
        for pos, kmer_idx in enumerate(self._get_kmer_bin_indexes(seq)):
            start_index = self.index_ranges[kmer_idx]
            end_index = self.index_ranges[kmer_idx + 1]
            for barcode_index in range(start_index, end_index):
                barcode = self.index[barcode_index]
                barcode_counts[barcode] += 1
                barcode_positions[barcode].append(pos)

        result = []
        for barcode in barcode_counts.keys():
            count = barcode_counts[barcode]
            if count < min_kmers:
                continue
            if ignore_equal and barcode == seq:
                continue
            result.append((barcode, count, barcode_positions[barcode]))

        if not result:
            return []

        top_hits = max(result, key=lambda x: x[1])[1]
        result = filter(lambda x: x[1] >= top_hits - hits_delta, result)
        result = sorted(result, reverse=True, key=lambda x: x[1])

        # Convert 2-bit sequences back to strings
        if max_hits == 0:
            return [(bit_to_str(x[0], self.seq_len), x[1], x[2]) for x in result]
        return [(bit_to_str(x[0], self.seq_len), x[1], x[2]) for x in list(result)[:max_hits]]



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import os
import pytest
import gffutils
from src.gene_info import *


class TestGeneInfo:
    source_dir = os.path.dirname(os.path.realpath(__file__))
    gffutils_db = gffutils.FeatureDB(os.path.join(source_dir, 'toy_data/synth.db'), keep_order=True)
    gene_db = gffutils_db['ENSMUSG00000020196.10']

    def test_basic(self):
        gene_info = GeneInfo([self.gene_db], self.gffutils_db)
        assert gene_info.get_gene_region() == ("chr10", 1000, 10000)
        assert gene_info.transcript_start("ENSMUST00000001712.7") == 1000
        assert gene_info.transcript_end("ENSMUST00000001713.7") == 10000
        assert gene_info.transcript_region("ENSMUST00000001715.7") == (8000, 8800)
        assert gene_info.transcript_exon_count("ENSMUST00000001713.7") == 6
        assert gene_info.total_transcript_length("ENSMUST00000001712.7") == 1105
        assert gene_info.chr_id == "chr10"
        assert gene_info.isoform_strands["ENSMUST00000001713.7"] == '-'
        assert gene_info.gene_id_map["ENSMUST00000001713.7"] == 'ENSMUSG00000020196.10'

    def test_intron_profiles(self):
        gene_info = GeneInfo([self.gene_db], self.gffutils_db)
        assert gene_info.intron_profiles.features[0] == (1101, 1999)
        assert gene_info.intron_profiles.features[-1] == (8201, 8499)
        assert len(gene_info.intron_profiles.features) == 9
        assert gene_info.intron_profiles.profiles["ENSMUST00000001712.7"] == [1, 1, -1, 1, -1, -1, 1, -1, -1]
        assert gene_info.intron_profiles.profiles["ENSMUST00000001714.7"] == [-2, -2, -2, -2, -2, -2, -1, -1, -2]
        assert gene_info.intron_profiles.profiles["ENSMUST00000001715.7"] == [-2, -2, -2, -2, -2, -2, -1, -1, 1]

    def test_exon_profiles(self):
        gene_info = GeneInfo([self.gene_db], self.gffutils_db)
        assert gene_info.exon_profiles.features[0] == (1000, 1100)
        assert gene_info.exon_profiles.features[-1] == (9500, 10000)
        assert len(gene_info.exon_profiles.features) == 11
        assert gene_info.exon_profiles.profiles["ENSMUST00000001712.7"] == [1, 1, -1, 1, -1, 1, -1, -1, -1, -1, 1]
        assert gene_info.exon_profiles.profiles["ENSMUST00000001714.7"] == [-2, -2, -2, -2, -2, -2, -2, 1, -2, -2, -2]

        assert gene_info.split_exon_profiles.features[2] == (2101, 2200)
        assert gene_info.split_exon_profiles.features[-1] == (9500, 10000)
        assert len(gene_info.split_exon_profiles.features) == 11
        assert gene_info.split_exon_profiles.profiles["ENSMUST00000001712.7"] == [1, 1, -1, 1, -1, 1, -1, -1, -1, -1, 1]
        assert gene_info.split_exon_profiles.profiles["ENSMUST00000001714.7"] == [-2, -2, -2, -2, -2, -2, -2, 1, -2, -2,
                                                                                  -2]

    def test_from_region(self):
        chr_id = "chr10"
        start, end = 1000, 2000
        gene_info = GeneInfo.from_region(chr_id, start, end)

        assert gene_info.chr_id == chr_id
        assert gene_info.start == start
        assert gene_info.end == end
        assert gene_info.db is None
        assert gene_info.gene_db_list == []
        assert gene_info.empty()

    def test_from_models(self):
        from src.gene_info import TranscriptModel

        # Create a simple transcript model
        transcript_model = TranscriptModel("chr1", "+", "test_transcript", "test_gene", [(100, 200), (300, 400)], TranscriptModelType.known)

        gene_info = GeneInfo.from_models([transcript_model])

        assert gene_info.chr_id == "chr1"
        assert gene_info.start == 100
        assert gene_info.end == 400
        assert gene_info.gene_id_map["test_transcript"] == "test_gene"
        assert gene_info.isoform_strands["test_transcript"] == "+"
        assert gene_info.all_isoforms_exons["test_transcript"] == [(100, 200), (300, 400)]

    def test_serialization(self):
        import io
        from src.serialization import write_int, write_string

        # Create test gene info
        gene_info = GeneInfo([self.gene_db], self.gffutils_db)

        # Serialize
        outfile = io.BytesIO()
        gene_info.serialize(outfile)
        outfile.seek(0)

        # Deserialize
        new_gene_info = GeneInfo.deserialize(outfile, self.gffutils_db)

        # Check key attributes
        assert new_gene_info.chr_id == gene_info.chr_id
        assert new_gene_info.start == gene_info.start
        assert new_gene_info.end == gene_info.end
        assert new_gene_info.delta == gene_info.delta

    def test_get_gene_regions(self):
        gene_info = GeneInfo([self.gene_db], self.gffutils_db)
        regions = gene_info.get_gene_regions()

        assert isinstance(regions, dict)
        assert self.gene_db.id in regions
        assert regions[self.gene_db.id] == (self.gene_db.start, self.gene_db.end)

    def test_empty(self):
        gene_info = GeneInfo([self.gene_db], self.gffutils_db)
        assert not gene_info.empty()

        # Test empty gene info
        empty_gene_info = GeneInfo.from_region("chr1", 1, 100)
        assert empty_gene_info.empty()

    def test_feature_properties(self):
        gene_info = GeneInfo([self.gene_db], self.gffutils_db)

        # Test exon properties
        exon_props = gene_info.exon_property_map
        assert len(exon_props) == len(gene_info.exon_profiles.features)
        for prop in exon_props:
            assert isinstance(prop, FeatureInfo)
            assert prop.chr_id == gene_info.chr_id
            assert len(prop.gene_ids) > 0

        # Test intron properties
        intron_props = gene_info.intron_property_map
        assert len(intron_props) == len(gene_info.intron_profiles.features)
        for prop in intron_props:
            assert isinstance(prop, FeatureInfo)
            assert prop.chr_id == gene_info.chr_id
            assert len(prop.gene_ids) > 0



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

# serialization stuff
ENCODING = 'utf-8'
BYTE_ORDER = "big"
STR_LEN_BYTES = 4
NONE_STR_LEN = (1 << (8 * STR_LEN_BYTES)) - 1
SHORT_INT_BYTES = 2
LONG_INT_BYTES = 4
TERMINATION_INT = (1 << 32) - 1
SHORT_TERMINATION_INT = (1 << 16) - 1
SHORT_FLOAT_MULTIPLIER = 1 << 20
DICT_TYPE_LEN = 1
DICT_INT_TYPE = 9
DICT_INT_PAIR_TYPE = 10
DICT_STR_TYPE = 17


def write_string(s, outf):
    str_len = len(s)
    outf.write(str_len.to_bytes(STR_LEN_BYTES, BYTE_ORDER) + bytearray(s, encoding=ENCODING))


def read_string(inf):
    str_len = int.from_bytes(inf.read(STR_LEN_BYTES), BYTE_ORDER)
    return inf.read(str_len).decode(encoding=ENCODING)


def write_string_or_none(s, outf):
    if s is None:
        outf.write(NONE_STR_LEN.to_bytes(STR_LEN_BYTES, BYTE_ORDER))
        return
    str_len = len(s)
    outf.write(str_len.to_bytes(STR_LEN_BYTES, BYTE_ORDER) + bytearray(s, encoding=ENCODING))


def read_string_or_none(inf):
    str_len = int.from_bytes(inf.read(STR_LEN_BYTES), BYTE_ORDER)
    if str_len == NONE_STR_LEN:
        return None
    return inf.read(str_len).decode(encoding=ENCODING)


def write_int_or_none(val, outf, bytes_len=LONG_INT_BYTES):
    if val is None:
        outf.write(TERMINATION_INT.to_bytes(bytes_len, BYTE_ORDER))
        return
    outf.write(val.to_bytes(bytes_len, BYTE_ORDER))


def read_int_or_none(inf, bytes_len=LONG_INT_BYTES):
    val = int.from_bytes(inf.read(bytes_len), BYTE_ORDER)
    if val == TERMINATION_INT:
        return None
    return val


def write_int(val, outf, bytes_len=LONG_INT_BYTES):
    outf.write(val.to_bytes(bytes_len, BYTE_ORDER))


def read_int(inf, bytes_len=LONG_INT_BYTES):
    return int.from_bytes(inf.read(bytes_len), BYTE_ORDER)


def write_short_int(val, outf):
    write_int(val, outf, SHORT_INT_BYTES)


def read_short_int(inf):
    return read_int(inf, SHORT_INT_BYTES)


def write_list(l, outf, func):
    write_int(len(l), outf)
    for val in l:
        func(val, outf)


def read_list(inf, func):
    result = []
    list_size = read_int(inf)
    for i in range(list_size):
        result.append(func(inf))
    return result


def write_list_of_pairs(l, outf, func):
    write_int(len(l), outf)
    for val in l:
        func(val[0], outf)
        func(val[1], outf)


def read_list_of_pairs(inf, func):
    result = []
    list_size = read_int(inf)
    for i in range(list_size):
        result.append((func(inf), func(inf)))
    return result


def write_bool_array(bool_arr, outf):
    assert len(bool_arr) <= 8
    byte_val = 0
    for i, val in enumerate(bool_arr):
        if val:
            byte_val |= 1 << i
    outf.write(byte_val.to_bytes(1, BYTE_ORDER))


def read_bool_array(inf, arr_size=1):
    byte_val = int.from_bytes(inf.read(1), BYTE_ORDER)
    bool_arr = []
    for i in range(arr_size):
        mask = 1 << i
        bool_arr.append(byte_val & mask != 0)
    return bool_arr


# transforms a signed integer into an unsigned integer while using the 31st bit as a flag to remember the original sign
def write_int_neg(val, outf):
    if val < 0:
        val = abs(val)
        assert val & 1 << 31 == 0
        val |= 1 << 31
    else:
        assert val & 1 << 31 == 0
    outf.write(val.to_bytes(LONG_INT_BYTES, BYTE_ORDER))


def read_int_neg(inf):
    val = int.from_bytes(inf.read(LONG_INT_BYTES), BYTE_ORDER)
    if val & 1 << 31 != 0:
        return - (val & ((1 << 31) - 1))
    return val


def write_dict(d, outf):
    # keys - only strings, values - string, ints, int pairs
    write_int(len(d), outf)
    for k in d.keys():
        write_string(k, outf)
        v = d[k]
        if isinstance(v, int):
            outf.write(DICT_INT_TYPE.to_bytes(DICT_TYPE_LEN, BYTE_ORDER))
            write_int_neg(v, outf)
        elif isinstance(v, str):
            outf.write(DICT_STR_TYPE.to_bytes(DICT_TYPE_LEN, BYTE_ORDER))
            write_string(v, outf)
        elif isinstance(v, tuple):
            outf.write(DICT_INT_PAIR_TYPE.to_bytes(DICT_TYPE_LEN, BYTE_ORDER))
            write_int_neg(v[0], outf)
            write_int_neg(v[1], outf)
        else:
            raise ValueError("Unsupported value type in dictionary serialization")


def read_dict(inf):
    d = {}
    d_len = read_int(inf)
    for i in range(d_len):
        k = read_string(inf)
        vtype = int.from_bytes(inf.read(DICT_TYPE_LEN), BYTE_ORDER)
        if vtype == DICT_INT_TYPE:
            d[k] = read_int(inf)
        elif vtype == DICT_STR_TYPE:
            d[k] = read_string(inf)
        elif vtype == DICT_INT_PAIR_TYPE:
            d[k] = (read_int(inf), read_int(inf))
        else:
            raise ValueError("Serialized dictionary contains unsupported value")
    return d



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Core barcode detection functionality.

This module contains the main barcode calling logic used by both
isoquant_detect_barcodes.py (standalone CLI) and isoquant.py (integrated pipeline).
"""

import concurrent.futures
import gc
import gzip
import logging
import multiprocessing
import os
import random
import shutil
import sys
from collections import defaultdict

import pysam
from Bio import SeqIO, Seq, SeqRecord

from ..modes import IsoQuantMode
from ..error_codes import IsoQuantExitCode
from ..common import setup_worker_logging, _get_log_params
from .common import reverese_complement, load_barcodes
from . import (
    TenXBarcodeDetector,
    TenXv2BarcodeDetector,
    CurioBarcodeDetector,
    SharedMemoryStereoBarcodeDetector,
    SharedMemoryStereoSplittingBarcodeDetector,
    ReadStats,
    VisiumHDBarcodeDetector,
    UniversalSingleMoleculeExtractor,
    MoleculeStructure
)

logger = logging.getLogger('IsoQuant')

READ_CHUNK_SIZE = 100000

BARCODE_CALLING_MODES = {
    IsoQuantMode.tenX_v3: TenXBarcodeDetector,
    IsoQuantMode.tenX_v2: TenXv2BarcodeDetector,
    IsoQuantMode.curio: CurioBarcodeDetector,
    IsoQuantMode.stereoseq_nosplit: SharedMemoryStereoBarcodeDetector,
    IsoQuantMode.stereoseq: SharedMemoryStereoSplittingBarcodeDetector,
    IsoQuantMode.visium_5prime: TenXBarcodeDetector,
    IsoQuantMode.visium_hd: VisiumHDBarcodeDetector,
    IsoQuantMode.custom_sc: UniversalSingleMoleculeExtractor
}

BARCODE_FILES_REQUIRED = {
    IsoQuantMode.tenX_v3: [1],
    IsoQuantMode.tenX_v2: [1],
    IsoQuantMode.curio: [1, 2],
    IsoQuantMode.stereoseq_nosplit: [1],
    IsoQuantMode.stereoseq: [1],
    IsoQuantMode.visium_5prime: [1],
    IsoQuantMode.visium_hd: [2],
    IsoQuantMode.custom_sc: [0]
}


def stats_file_name(file_name):
    return file_name + ".stats"


def get_umi_length(isoquant_mode: IsoQuantMode):
    if isoquant_mode not in BARCODE_CALLING_MODES:
        return 0
    try:
        return BARCODE_CALLING_MODES[isoquant_mode].UMI_LEN
    except AttributeError:
        return 0


class SimpleReadStorage:
    def __init__(self):
        self.read_ids = []
        self.sequences = []

    def add(self, read_id, seq):
        self.read_ids.append(read_id)
        self.sequences.append(seq)

    def clear(self):
        self.read_ids.clear()
        self.sequences.clear()

    def __len__(self):
        return len(self.read_ids)

    def __iter__(self):
        for i in range(len(self.read_ids)):
            yield self.read_ids[i], self.sequences[i]

    def __getstate__(self):
        return self.read_ids, self.sequences

    def __setstate__(self, state):
        self.read_ids = state[0]
        self.sequences = state[1]


class BarcodeCaller:
    def __init__(self, output_file_name, barcode_detector, header=False, output_sequences=None):
        self.barcode_detector = barcode_detector
        self.output_file_name = output_file_name
        self.output_file = open(self.output_file_name, "w")
        self.output_sequences = output_sequences
        self.output_sequences_file = None
        self.process_function = self._process_read_normal
        if self.output_sequences:
            self.output_sequences_file = open(self.output_sequences, "w")
            self.process_function = self._process_read_split
        if header:
            self.output_file.write(barcode_detector.header() + "\n")
        self.read_stat = ReadStats()

    def get_stats(self):
        return self.read_stat

    def dump_stats(self, file_name=None):
        if not file_name:
            file_name = stats_file_name(self.output_file_name)
        stat_out = open(file_name, "w")
        stat_out.write(str(self.read_stat))
        stat_out.close()

    def close(self):
        self.output_file.close()
        if self.output_sequences_file:
            self.output_sequences_file.close()

    def __del__(self):
        if not self.output_file.closed:
            self.output_file.close()
        if self.output_sequences_file and not self.output_sequences_file.closed:
            self.output_sequences_file.close()

    def process(self, input_file):
        logger.info("Processing " + input_file)
        fname, outer_ext = os.path.splitext(os.path.basename(input_file))
        low_ext = outer_ext.lower()

        handle = input_file
        if low_ext in ['.gz', '.gzip']:
            handle = gzip.open(input_file, "rt")
            input_file = fname
            fname, outer_ext = os.path.splitext(os.path.basename(input_file))
            low_ext = outer_ext.lower()

        if low_ext in ['.fq', '.fastq']:
            self._process_fastx(SeqIO.parse(handle, "fastq"))
        elif low_ext in ['.fa', '.fasta']:
            self._process_fastx(SeqIO.parse(handle, "fasta"))
        elif low_ext in ['.bam', '.sam']:
            self._process_bam(pysam.AlignmentFile(input_file, "rb", check_sq=False))
        else:
            logger.error("Unknown file format " + input_file)

        logger.info("Finished " + input_file)

    def _process_fastx(self, read_handler):
        counter = 0
        for r in read_handler:
            if counter % 100 == 0:
                sys.stdout.write("Processed %d reads\r" % counter)
            counter += 1
            read_id = r.id
            seq = str(r.seq)
            self.process_function(read_id, seq)

    def _process_bam(self, read_handler):
        counter = 0
        for r in read_handler:
            if counter % 100 == 0:
                sys.stdout.write("Processed %d reads\r" % counter)
            counter += 1
            read_id = r.query_name
            seq = r.query_sequence
            self.process_function(read_id, seq)

    # split read and find multiple barcodes
    def _process_read_split(self, read_id, read_sequence):
        logger.debug("==== %s ====" % read_id)
        barcode_result = self.barcode_detector.find_barcode_umi(read_id, read_sequence)

        seq_records = []
        require_tso = len(barcode_result.detected_patterns) > 1
        strands = set()
        for r in barcode_result.detected_patterns:
            self.read_stat.add_read(r)
            if not r.is_valid():
                self.output_file.write("%s\n" % str(r))
                continue

            read_segment_start = max(0, r.primer - 25, r.polyT - 75)
            read_segment_end = len(read_sequence) if r.tso5 == -1 else min(len(read_sequence), r.tso5 + 25)
            r.read_id = read_id + ("_%d_%d_%s" % (read_segment_start, read_segment_end, r.strand))
            if r.strand == "+":
                new_read_seq = read_sequence[read_segment_start:read_segment_end]
            else:
                new_read_seq = reverese_complement(read_sequence)[read_segment_start:read_segment_end]
            strands.add(r.strand)
            self.output_file.write("%s\n" % str(r))
            if self.output_sequences and (not require_tso or r.tso5 != -1):
                seq_records.append(SeqRecord.SeqRecord(seq=Seq.Seq(new_read_seq), id=r.read_id, description=""))

        self.read_stat.add_custom_stats("Splits", len(barcode_result.detected_patterns))
        if self.output_sequences_file:
            SeqIO.write(seq_records, self.output_sequences_file, "fasta")

    def _process_read_normal(self, read_id, read_sequence):
        logger.debug("==== %s ====" % read_id)
        if read_sequence is None:
            return
        barcode_result = self.barcode_detector.find_barcode_umi(read_id, read_sequence)

        self.output_file.write("%s\n" % str(barcode_result))
        self.read_stat.add_read(barcode_result)

    def process_chunk(self, read_chunk):
        counter = 0
        for read_id, seq in read_chunk:
            self.process_function(read_id, seq)
            counter += 1
        return counter


def fastx_file_chunk_reader(handler):
    current_chunk = SimpleReadStorage()
    for r in handler:
        current_chunk.add(r.id, str(r.seq))
        if len(current_chunk) >= READ_CHUNK_SIZE:
            yield current_chunk
            current_chunk = SimpleReadStorage()
    yield current_chunk


def bam_file_chunk_reader(handler):
    current_chunk = SimpleReadStorage()
    for r in handler:
        if r.is_secondary or r.is_supplementary:
            continue
        current_chunk.add(r.query_name, r.query_sequence)
        if len(current_chunk) >= READ_CHUNK_SIZE:
            yield current_chunk
            current_chunk = SimpleReadStorage()
    yield current_chunk


def process_chunk(barcode_detector, read_chunk, output_file, num, out_fasta=None):
    output_file += "_" + str(num)
    if out_fasta:
        out_fasta += "_" + str(num)
    counter = 0

    barcode_caller = BarcodeCaller(output_file, barcode_detector, output_sequences=out_fasta)
    counter += barcode_caller.process_chunk(read_chunk)
    read_chunk.clear()
    barcode_caller.dump_stats()
    barcode_caller.close()

    return output_file, out_fasta, counter


def create_barcode_caller(args):
    logger.info("Creating barcode detector for mode %s" % args.mode.name)

    if args.mode == IsoQuantMode.custom_sc:
        if not args.molecule:
            logger.critical("Custom single-cell/spatial mode requires molecule description to be provided via --molecule")
            sys.exit(IsoQuantExitCode.INCOMPATIBLE_OPTIONS)
        if not os.path.isfile(args.molecule):
            logger.critical("Molecule file %s does not exist" % args.molecule)
            sys.exit(IsoQuantExitCode.INPUT_FILE_NOT_FOUND)

        return BARCODE_CALLING_MODES[args.mode](MoleculeStructure(open(args.molecule)))

    logger.info("Using barcodes from %s" % ", ".join(args.barcodes))
    barcode_files = len(args.barcodes)
    if barcode_files not in BARCODE_FILES_REQUIRED[args.mode]:
        logger.critical("Barcode calling mode %s requires %s files, %d provided" %
                        (args.mode.name, " or ".join([str(x) for x in BARCODE_FILES_REQUIRED[args.mode]]), barcode_files))
        sys.exit(IsoQuantExitCode.BARCODE_FILE_COUNT_MISMATCH)

    barcodes = []
    for bc in args.barcodes:
        barcodes.append(load_barcodes(bc, needs_iterator=args.mode.needs_barcode_iterator()))

    if len(barcodes) == 1:
        barcodes = barcodes[0]
        if not args.mode.needs_barcode_iterator():
            logger.info("Loaded %d barcodes" % len(barcodes))
    else:
        if not args.mode.needs_barcode_iterator():
            for i, bc in enumerate(barcodes):
                logger.info("Loaded %d barcodes from %s" % (len(bc), args.barcodes[i]))
        barcodes = tuple(barcodes)

    barcode_detector = BARCODE_CALLING_MODES[args.mode](barcodes)

    return barcode_detector


def process_single_thread(args):
    barcode_detector = create_barcode_caller(args)

    # args.input, args.output_tsv are always lists
    # args.out_fasta is a list or None
    out_fastas = args.out_fasta if args.out_fasta else [None] * len(args.input)

    for idx, (input_file, output_tsv, out_fasta) in enumerate(zip(args.input, args.output_tsv, out_fastas)):
        if len(args.input) > 1:
            logger.info("Processing file %d/%d: %s" % (idx + 1, len(args.input), input_file))
        barcode_caller = BarcodeCaller(output_tsv, barcode_detector, header=True, output_sequences=out_fasta)
        barcode_caller.process(input_file)
        barcode_caller.dump_stats()
        for stat_line in barcode_caller.get_stats():
            logger.info("  " + stat_line)
        barcode_caller.close()
    logger.info("Finished barcode calling")


def _process_single_file_in_parallel(input_file, output_tsv, out_fasta, args, barcode_detector):
    """Process a single file in parallel (internal helper function)."""
    logger.info("Processing " + input_file)
    fname, outer_ext = os.path.splitext(os.path.basename(input_file))
    low_ext = outer_ext.lower()

    handle = input_file
    if low_ext in ['.gz', '.gzip']:
        handle = gzip.open(input_file, "rt")
        input_file = fname
        fname, outer_ext = os.path.splitext(os.path.basename(input_file))
        low_ext = outer_ext.lower()

    if low_ext in ['.fq', '.fastq']:
        read_chunk_gen = fastx_file_chunk_reader(SeqIO.parse(handle, "fastq"))
    elif low_ext in ['.fa', '.fasta']:
        read_chunk_gen = fastx_file_chunk_reader(SeqIO.parse(handle, "fasta"))
    elif low_ext in ['.bam', '.sam']:
        read_chunk_gen = bam_file_chunk_reader(pysam.AlignmentFile(input_file, "rb", check_sq=False))
    else:
        logger.error("Unknown file format " + input_file)
        sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)

    tmp_dir = "barcode_calling_%x" % random.randint(0, 1 << 32)
    while os.path.exists(tmp_dir):
        tmp_dir = "barcode_calling_%x" % random.randint(0, 1 << 32)
    if args.tmp_dir:
        tmp_dir = os.path.join(args.tmp_dir, tmp_dir)
    else:
        tmp_dir = os.path.join(os.path.dirname(args.output), tmp_dir)
    os.makedirs(tmp_dir)

    tmp_barcode_file = os.path.join(tmp_dir, "bc")
    tmp_fasta_file = os.path.join(tmp_dir, "subreads") if out_fasta else None
    chunk_counter = 0
    future_results = []
    output_files = []

    # Clean up parent memory before spawning workers
    gc.collect()
    mp_context = multiprocessing.get_context('spawn')
    log_file, log_level = _get_log_params()
    # max_tasks_per_child requires Python 3.11+
    executor_kwargs = {
        'max_workers': args.threads,
        'mp_context': mp_context,
        'initializer': setup_worker_logging,
        'initargs': (log_file, log_level),
    }
    if sys.version_info >= (3, 11):
        executor_kwargs['max_tasks_per_child'] = 20
    with concurrent.futures.ProcessPoolExecutor(**executor_kwargs) as proc:
        for chunk in read_chunk_gen:
            future_results.append(proc.submit(process_chunk,
                                              barcode_detector,
                                              chunk,
                                              tmp_barcode_file,
                                              chunk_counter,
                                              tmp_fasta_file))
            chunk_counter += 1
            if chunk_counter >= args.threads:
                break

        reads_left = True
        read_counter = 0
        while future_results:
            completed_features, _ = concurrent.futures.wait(future_results,
                                                            return_when=concurrent.futures.FIRST_COMPLETED)
            for c in completed_features:
                if c.exception() is not None:
                    raise c.exception()
                res = c.result()
                tmp_out_file, tmp_out_fasta, read_count = res
                read_counter += read_count
                sys.stdout.write("Processed %d reads\r" % read_counter)
                output_files.append((tmp_out_file, tmp_out_fasta))
                future_results.remove(c)
                if reads_left:
                    try:
                        chunk = next(read_chunk_gen)
                        future_results.append(proc.submit(process_chunk,
                                                          barcode_detector,
                                                          chunk,
                                                          tmp_barcode_file,
                                                          chunk_counter,
                                                          tmp_fasta_file))
                        chunk_counter += 1
                    except StopIteration:
                        reads_left = False

    with open(output_tsv, "w") as final_output_tsv:
        final_output_fasta = open(out_fasta, "w") if out_fasta else None
        header = barcode_detector.header()
        final_output_tsv.write(header + "\n")
        stat_dict = defaultdict(int)
        for tmp_file, tmp_fasta in output_files:
            shutil.copyfileobj(open(tmp_file, "r"), final_output_tsv)
            if tmp_fasta and final_output_fasta:
                shutil.copyfileobj(open(tmp_fasta, "r"), final_output_fasta)
            for l in open(stats_file_name(tmp_file), "r"):
                v = l.strip().split("\t")
                if len(v) != 2:
                    continue
                stat_dict[v[0]] += int(v[1])

        if final_output_fasta is not None:
            final_output_fasta.close()

    with open(stats_file_name(output_tsv), "w") as out_stats:
        for k, v in stat_dict.items():
            logger.info("  %s: %d" % (k, v))
            out_stats.write("%s\t%d\n" % (k, v))
    shutil.rmtree(tmp_dir)


def process_in_parallel(args):
    """Process input files in parallel."""
    # args.input, args.output_tsv are always lists
    # args.out_fasta is a list or None
    out_fastas = args.out_fasta if args.out_fasta else [None] * len(args.input)
    barcode_detector = create_barcode_caller(args)

    for idx, (input_file, output_tsv, out_fasta) in enumerate(zip(args.input, args.output_tsv, out_fastas)):
        if len(args.input) > 1:
            logger.info("Processing file %d/%d: %s" % (idx + 1, len(args.input), input_file))
        _process_single_file_in_parallel(input_file, output_tsv, out_fasta, args, barcode_detector)

    logger.info("Finished barcode calling")



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
from functools import partial

from .common import equal_ranges, junctions_from_blocks, contains_well_inside
from .isoform_assignment import ReadAssignmentType, MatchEventSubtype, SupplementaryMatchConstants
from .long_read_profiles import OverlappingFeaturesProfileConstructor

logger = logging.getLogger('IsoQuant')


class ExonCorrector:
    def __init__(self, gene_info, params, chr_record):
        self.gene_info = gene_info
        self.chr_record = chr_record
        self.params = params
        self.delta = self.params.delta
        self.intron_profile_constructor = \
            OverlappingFeaturesProfileConstructor(self.gene_info.intron_profiles.features,
                                                  (self.gene_info.start, self.gene_info.end),
                                                  comparator=partial(equal_ranges, delta=self.params.delta))

    def correct_assigned_read(self, alignment_info, read_assignment):
        if len(alignment_info.read_exons) == 1 or \
                read_assignment.assignment_type == ReadAssignmentType.noninformative or \
                not read_assignment.isoform_matches:
            return alignment_info.read_exons

        read_region, new_introns = self.correct_misalignments(alignment_info, read_assignment)
        if new_introns:
            corrected_exons = [(read_region[0], new_introns[0][0] - 1)]
            corrected_exons += junctions_from_blocks(new_introns)
            corrected_exons += [(new_introns[-1][1] + 1, read_region[1])]
        else:
            corrected_exons = [(read_region[0], read_region[1])]
        return corrected_exons

    def correct_fuzzy_junctions(self, alignment_info, read_assignment):
        intron_profile = alignment_info.combined_profile.read_intron_profile
        isoform_id = read_assignment.isoform_matches[0].assigned_transcript
        isofrom_profile = self.gene_info.intron_profiles.profiles[isoform_id]
        ref_introns = self.gene_info.intron_profiles.features

        # logger.debug("Correcting fuzzy junctions in the following profiles")
        # logger.debug(intron_profile.read_profile)
        # logger.debug(intron_profile.gene_profile)
        # logger.debug(isofrom_profile)

        new_introns = []
        read_pos, ref_pos = self.find_next(intron_profile, -1, -1)
        while read_pos < len(intron_profile.read_profile) and ref_pos < len(intron_profile.gene_profile):
            if isofrom_profile[ref_pos] != 1:
                # if intron match is ambiguous, we have to choose the one that belongs to the assigned isoform
                ref_pos += 1
            assert isofrom_profile[ref_pos] == 1

            read_intron = intron_profile.read_features[read_pos]
            ref_intron = ref_introns[ref_pos]
            assert equal_ranges(ref_intron, read_intron, self.delta)
            # TODO: check for reliability of splice sites
            new_introns.append(ref_intron)
            read_pos, ref_pos = self.find_next(intron_profile, read_pos, ref_pos)
        return (alignment_info.read_start, alignment_info.read_end), new_introns

    def find_next(self, intron_profile, read_pos, ref_pos):
        try:
            # first matched read intron
            read_pos = intron_profile.read_profile.index(1, read_pos + 1)
            # first matched isoform intron
            ref_pos = intron_profile.gene_profile.index(1, ref_pos + 1)
        except ValueError:
            read_pos = len(intron_profile.read_profile)
            ref_pos = len(intron_profile.gene_profile)
        return read_pos, ref_pos

    def correct_misalignments(self, alignment_info, read_assignment):
        event_map = {}
        for e in read_assignment.isoform_matches[0].match_subclassifications:
            if e.read_region == SupplementaryMatchConstants.undefined_region:
                continue
            if e.read_region[0] == SupplementaryMatchConstants.absent_position:
                if e.event_type == MatchEventSubtype.fake_micro_intron_retention and \
                        self.params.correct_microintron_retention:
                    event_map[-e.read_region[1]-1] = e
            else:
                event_map[e.read_region[0]] = e

        intron_profile = alignment_info.combined_profile.read_intron_profile
        read_introns = intron_profile.read_features
        read_region = (alignment_info.read_start, alignment_info.read_end)
        isoform_id = read_assignment.isoform_matches[0].assigned_transcript
        isoform_region = self.gene_info.transcript_region(isoform_id)
        isoform_introns = self.gene_info.all_isoforms_introns[isoform_id]
        return self.process_events(alignment_info, event_map, read_region, read_introns,  isoform_region, isoform_introns)

    def process_events(self, alignment_info, event_map, read_region, read_introns, isoform_region, isoform_introns):
        if self.params.correct_fuzzy_junctions:
            # logger.debug("*** Correcting fuzzy junctions")
            potential_introns = self.intron_profile_constructor.match_genomic_features(read_introns)
            # logger.debug(potential_introns)
            corrected_introns = []
            assert len(potential_introns) == len(read_introns)
            for i in range(len(read_introns)):
                read_intron = read_introns[i]
                ref_intron = potential_introns[i]

                if read_intron[0] == ref_intron[0]:
                    left_site = read_intron[0]
                    # logger.debug("Equal left sites %d" % read_intron[0])
                else:
                    start = min(read_intron[0], ref_intron[0])
                    end = max(read_intron[0], ref_intron[0]) - 1
                    # logger.debug("Unequal left sites %d - %d" % (start, end))
                    indel_count, mm_count = alignment_info.get_error_count(start, end, intron_index=i,
                                                                           left_site=True, chr_record=self.chr_record)
                    # TODO: check condition
                    left_site = read_intron[0] if indel_count == 0 and mm_count <= 1 else ref_intron[0]
                    # logger.debug("Errors: %d, %d; res: %d" % (indel_count, mm_count, left_site))

                if read_intron[1] == ref_intron[1]:
                    right_site = read_intron[1]
                    # logger.debug("Equal right sites %d" % right_site)
                else:
                    start = min(read_intron[1], ref_intron[1]) + 1
                    end = max(read_intron[1], ref_intron[1])
                    # logger.debug("Unequal right sites %d - %d" % (start, end))
                    indel_count, mm_count = alignment_info.get_error_count(start, end, intron_index=i,
                                                                           left_site=False, chr_record=self.chr_record)
                    # TODO: check condition
                    right_site = read_intron[1] if indel_count == 0 and mm_count <= 1 else ref_intron[1]
                    # logger.debug("Errors: %d, %d; res: %d" % (indel_count, mm_count, left_site))

                corrected_introns.append((left_site, right_site))
        else:
            corrected_introns = read_introns

        assert len(corrected_introns) == len(read_introns)
        # logger.debug(read_introns)
        # logger.debug(corrected_introns)

        corrected_read_region = read_region
        new_introns = []
        i = 0

        while i < len(corrected_introns):
            if -i-1 in event_map:
                # special case for fake IR
                # logger.debug(event_map[-i-1].isoform_region)
                new_introns.append(isoform_introns[event_map[-i-1].isoform_region[0]])

            if i not in event_map:
                # TODO: check for reliability of splice sites
                new_introns.append(corrected_introns[i])
                i += 1
                continue

            misalignment_set = []
            if self.params.correct_intron_shifts:
                misalignment_set.append(MatchEventSubtype.intron_shift)
            if self.params.correct_skipped_exons:
                misalignment_set.append(MatchEventSubtype.exon_misalignment)

            event = event_map[i]
            if event.event_type == MatchEventSubtype.fake_terminal_exon_left and \
                    self.params.correct_fake_terminal_exons:
                assert event.read_region[0] == event.read_region[1]
                # fake terminal exon, skip it
                corrected_read_region = (read_introns[event.read_region[0]][1]+1, corrected_read_region[1])
            elif event.event_type == MatchEventSubtype.fake_terminal_exon_right and \
                    self.params.correct_fake_terminal_exons:
                assert event.read_region[0] == event.read_region[1]
                # fake terminal exon, skip it
                corrected_read_region = (corrected_read_region[0], read_introns[event.read_region[0]][0]-1)
            elif event.event_type == MatchEventSubtype.terminal_exon_misalignment_left and \
                    self.params.correct_terminal_exons:
                new_introns.append(isoform_introns[event.isoform_region[0]])
                corrected_read_region = (isoform_region[0], corrected_read_region[1])
            elif event.event_type == MatchEventSubtype.terminal_exon_misalignment_right and \
                    self.params.correct_terminal_exons:
                new_introns.append(isoform_introns[event.isoform_region[0]])
                corrected_read_region = (corrected_read_region[0], isoform_region[1])
            elif event.event_type in misalignment_set and \
                    contains_well_inside(read_region, (isoform_introns[event.isoform_region[0]][0],
                                                       isoform_introns[event.isoform_region[1]][1]), self.params.delta):
                # misalignments but inside read region
                assert event.read_region[0] == event.read_region[1]
                new_introns += [isoform_introns[i] for i in range(event.isoform_region[0], event.isoform_region[1] + 1)]
            elif event.event_type in {MatchEventSubtype.extra_intron_known,
                                      MatchEventSubtype.intron_alternation_known,
                                      MatchEventSubtype.intron_migration,
                                      MatchEventSubtype.exon_skipping_known,
                                      MatchEventSubtype.exon_merge_known,
                                      MatchEventSubtype.terminal_exon_shift_known,
                                      MatchEventSubtype.mutually_exclusive_exons_known,
                                      MatchEventSubtype.exon_gain_known,
                                      MatchEventSubtype.exon_detach_known,
                                      MatchEventSubtype.alternative_structure_known,
                                      MatchEventSubtype.alternative_structure_novel}:
                # add corrected read intron
                new_introns += [corrected_introns[i] for i in range(event.read_region[0], event.read_region[1] + 1)]
            else:
                # add as is
                new_introns += [read_introns[i] for i in range(event.read_region[0], event.read_region[1] + 1)]

            i = event.read_region[1] + 1

        return corrected_read_region, new_introns


